import copy
import types

import torch
from torch import nn
from transformers.models.t5.modeling_t5 import T5ForConditionalGeneration, T5Stack

from .modeling_t5 import (
    T5ForConditionalGeneration as T5ConditionalGenerationCrossAttentionScore,
)
from .modeling_t5 import T5Stack as T5StackCrossAttentionScore


class FiDStack(T5Stack):
    def __init__(self, config, embed_tokens=None):
        super().__init__(config, embed_tokens=embed_tokens)
        self._n_passages = None

    def reset_n_passages(self, n_passages: int):
        self._n_passages = n_passages

    def forward(
        self,
        input_ids=None,
        attention_mask=None,
        encoder_hidden_states=None,
        encoder_attention_mask=None,
        inputs_embeds=None,
        head_mask=None,
        cross_attn_head_mask=None,
        past_key_values=None,
        use_cache=None,
        output_attentions=None,
        output_hidden_states=None,
        return_dict=None,
        cache_position=None,
    ):
        if not self.is_decoder:
            input_ids = input_ids.view(input_ids.size(0) * self._n_passages, -1)
            attention_mask = attention_mask.view(
                attention_mask.size(0) * self._n_passages, -1
            )

        output = super().forward(
            input_ids=input_ids,
            attention_mask=attention_mask,
            encoder_hidden_states=encoder_hidden_states,
            encoder_attention_mask=encoder_attention_mask,
            inputs_embeds=inputs_embeds,
            head_mask=head_mask,
            cross_attn_head_mask=cross_attn_head_mask,
            past_key_values=past_key_values,
            use_cache=use_cache,
            output_attentions=output_attentions,
            output_hidden_states=output_hidden_states,
            return_dict=return_dict,
            cache_position=cache_position,
        )

        if not self.is_decoder:
            bsz = input_ids.size(0) // self._n_passages
            if not return_dict:
                last_hidden_states = output[0]
                last_hidden_state = last_hidden_states.view(
                    bsz, -1, last_hidden_states.size(-1)
                )
                output = tuple(
                    last_hidden_state,
                    *output[1:],
                )
            else:
                last_hidden_state = output.last_hidden_state
                output.last_hidden_state = last_hidden_state.view(
                    bsz, -1, last_hidden_state.size(-1)
                )

        return output


class FiD(T5ForConditionalGeneration):
    _keys_to_ignore_on_load_missing = [
        r"encoder\.embed_tokens\.weight",
        r"decoder\.embed_tokens\.weight",
        r"lm_head\.weight",
    ]
    _keys_to_ignore_on_load_unexpected = [
        r"decoder\.block\.0\.layer\.1\.EncDecAttention\.relative_attention_bias\.weight",
    ]

    ANSWER_EOS_TOKEN = 1

    def __init__(self, config):
        super().__init__(config)
        self.model_dim = config.d_model

        self.shared = nn.Embedding(config.vocab_size, config.d_model)

        encoder_config = copy.deepcopy(config)
        encoder_config.is_decoder = False
        encoder_config.use_cache = False
        encoder_config.is_encoder_decoder = False

        self.encoder = FiDStack(encoder_config, self.shared)

        decoder_config = copy.deepcopy(config)
        decoder_config.is_decoder = True
        decoder_config.is_encoder_decoder = False
        decoder_config.use_cache = True
        decoder_config.num_layers = config.num_decoder_layers
        self.decoder = FiDStack(decoder_config, self.shared)

        self.lm_head = nn.Linear(config.d_model, config.vocab_size, bias=False)

        # Initialize weights and apply final processing
        self.post_init()

        # Model parallel
        self.model_parallel = False
        self.device_map = None

    def reset_n_passages(self, n_passages: int):
        self.encoder.reset_n_passages(n_passages)
        self.decoder.reset_n_passages(n_passages)

    def set_checkpoint(self, use_checkpoint):
        """
        Enable or disable checkpointing in the encoder.
        See https://pytorch.org/docs/stable/checkpoint.html
        """
        for mod in self.encoder.encoder.block:
            mod.use_checkpoint = use_checkpoint

    def reset_score_storage(self):
        """
        Reset score storage, only used when cross-attention scores are saved
        to train a retriever.
        """
        for mod in self.decoder.block:
            mod.layer[1].EncDecAttention.normalized_score_storage = None

    @torch.no_grad()
    def get_crossattention_scores(
        self, n_passages, mask, ids, mask_query=None, output_sequence_lengths=[]
    ):
        """
        Cross-attention scores are aggregated to obtain a single scalar per
        passage. This scalar can be seen as a similarity score between the
        question and the input passage. It is obtained by averaging the
        cross-attention scores obtained on the first decoded token over heads,
        layers, and tokens of the input passage.

        More details in Distilling Knowledge from Reader to Retriever:
        https://arxiv.org/abs/2012.04584.
        """
        norms = []
        for mod in self.decoder.block:
            norms.append(mod.layer[1].EncDecAttention.normalized_score_storage)
        norms = torch.stack(norms)

        output = {}
        self.aggregate_value(
            norms,
            mask,
            n_passages,
            ids,
            mask_query,
            output,
            prefix="norms",
            output_sequence_lengths=output_sequence_lengths,
        )
        return output

    def aggregate_value(
        self,
        scores,
        mask,
        n_passages,
        ids,
        mask_query=None,
        output={},
        prefix="",
        output_sequence_lengths=[],
    ):
        n_layers, bsz, n_tokens, total_tokens = scores.size()

        ids = ids.view(bsz, n_passages, -1)
        scores = scores.view(n_layers, bsz, n_tokens, n_passages, -1)
        mask = mask.view(bsz, n_passages, -1)
        scores = scores.masked_fill(~mask[None, :, None], 0.0)

        scores = scores.sum(dim=[0])

        scores_woquery = None
        # Compute scores based on scores without query
        if not mask_query is None:
            output[f"{prefix}woquery"] = self.get_woquery_score(
                scores,
                mask_query,
                mask,
                n_layers,
                output_sequence_lengths=output_sequence_lengths,
            )

        return output

    def get_woquery_score(
        self, scores, mask_query, mask, n_layers, output_sequence_lengths
    ):
        if scores.size(-1) > mask_query.size(-1):
            zero_padding = torch.zeros(
                [mask_query.size(0), scores.size(-1) - mask_query.size(-1)],
                device=mask_query.device,
                dtype=torch.bool,
            )
            mask_query = torch.cat([mask_query, zero_padding], dim=-1)
        mask_query = mask * (~mask_query[:, None])
        scores_woquery = scores.masked_fill(~mask_query[:, None], 0.0)

        ntokens_woquery = 256 * n_layers

        # zero out scores after EOS token. This is needed when batching results in sequences with different lengths.
        for i in range(len(scores_woquery)):
            scores_woquery[i, output_sequence_lengths[i] :, :, :] = 0

        scores_woquery = scores_woquery.sum(dim=[1, 3])
        return scores_woquery / ntokens_woquery

    def overwrite_forward_crossattention(self):
        """
        Replace cross-attention forward function, only used to save
        cross-attention scores.
        """
        for mod in self.decoder.block:
            xattn = mod.layer[1].EncDecAttention
            xattn.forward = types.MethodType(cross_attention_forward, xattn)

    def create_crossattention_storage(self):
        for mod in self.decoder.block:
            xattn = mod.layer[1].EncDecAttention
            xattn.normalized_score_storage = None


class FiDStackCrossAttentionScore(T5StackCrossAttentionScore):
    def __init__(self, config, embed_tokens=None):
        super().__init__(config, embed_tokens=embed_tokens)

    def forward(
        self,
        input_ids=None,
        attention_mask=None,
        encoder_hidden_states=None,
        encoder_attention_mask=None,
        inputs_embeds=None,
        head_mask=None,
        cross_attn_head_mask=None,
        past_key_values=None,
        use_cache=None,
        output_attentions=None,
        output_hidden_states=None,
        return_dict=None,
        n_passages: int = None,
    ):
        if not self.is_decoder:
            input_ids = input_ids.view(input_ids.size(0) * n_passages, -1)
            attention_mask = attention_mask.view(
                attention_mask.size(0) * n_passages, -1
            )

        output = super().forward(
            input_ids=input_ids,
            attention_mask=attention_mask,
            encoder_hidden_states=encoder_hidden_states,
            encoder_attention_mask=encoder_attention_mask,
            inputs_embeds=inputs_embeds,
            head_mask=head_mask,
            cross_attn_head_mask=cross_attn_head_mask,
            past_key_values=past_key_values,
            use_cache=use_cache,
            output_attentions=output_attentions,
            output_hidden_states=output_hidden_states,
            return_dict=return_dict,
        )

        if not self.is_decoder:
            bsz = input_ids.size(0) // n_passages
            if not return_dict:
                last_hidden_states = output[0]
                last_hidden_state = last_hidden_states.view(
                    bsz, -1, last_hidden_states.size(-1)
                )
                output = tuple(
                    last_hidden_state,
                    *output[1:],
                )
            else:
                last_hidden_state = output.last_hidden_state
                output.last_hidden_state = last_hidden_state.view(
                    bsz, -1, last_hidden_state.size(-1)
                )

        return output


class FiDCrossAttentionScore(T5ConditionalGenerationCrossAttentionScore):
    _keys_to_ignore_on_load_missing = [
        r"encoder\.embed_tokens\.weight",
        r"decoder\.embed_tokens\.weight",
        r"lm_head\.weight",
    ]
    _keys_to_ignore_on_load_unexpected = [
        r"decoder\.block\.0\.layer\.1\.EncDecAttention\.relative_attention_bias\.weight",
    ]

    ANSWER_EOS_TOKEN = 1

    def __init__(self, config):
        super().__init__(config)
        self.model_dim = config.d_model

        self.shared = nn.Embedding(config.vocab_size, config.d_model)

        encoder_config = copy.deepcopy(config)
        encoder_config.is_decoder = False
        encoder_config.use_cache = False
        encoder_config.is_encoder_decoder = False

        self.encoder = FiDStackCrossAttentionScore(encoder_config, self.shared)

        decoder_config = copy.deepcopy(config)
        decoder_config.is_decoder = True
        decoder_config.is_encoder_decoder = False
        decoder_config.use_cache = True
        decoder_config.num_layers = config.num_decoder_layers
        self.decoder = FiDStackCrossAttentionScore(decoder_config, self.shared)

        self.lm_head = nn.Linear(config.d_model, config.vocab_size, bias=False)

        # Initialize weights and apply final processing
        self.post_init()

        # Model parallel
        self.model_parallel = False
        self.device_map = None

    def set_checkpoint(self, use_checkpoint):
        """
        Enable or disable checkpointing in the encoder.
        See https://pytorch.org/docs/stable/checkpoint.html
        """
        for mod in self.encoder.encoder.block:
            mod.use_checkpoint = use_checkpoint

    def reset_score_storage(self):
        """
        Reset score storage, only used when cross-attention scores are saved
        to train a retriever.
        """
        for mod in self.decoder.block:
            mod.layer[1].EncDecAttention.normalized_score_storage = None

    @torch.no_grad()
    def get_crossattention_scores(
        self, n_passages, mask, ids, mask_query=None, output_sequence_lengths=[]
    ):
        """
        Cross-attention scores are aggregated to obtain a single scalar per
        passage. This scalar can be seen as a similarity score between the
        question and the input passage. It is obtained by averaging the
        cross-attention scores obtained on the first decoded token over heads,
        layers, and tokens of the input passage.

        More details in Distilling Knowledge from Reader to Retriever:
        https://arxiv.org/abs/2012.04584.
        """
        norms = []
        for mod in self.decoder.block:
            norms.append(mod.layer[1].EncDecAttention.normalized_score_storage)
        norms = torch.stack(norms)

        output = {}
        self.aggregate_value(
            norms,
            mask,
            n_passages,
            ids,
            mask_query,
            output,
            prefix="norms",
            output_sequence_lengths=output_sequence_lengths,
        )
        return output

    def aggregate_value(
        self,
        scores,
        mask,
        n_passages,
        ids,
        mask_query=None,
        output={},
        prefix="",
        output_sequence_lengths=[],
    ):
        n_layers, bsz, n_tokens, total_tokens = scores.size()

        ids = ids.view(bsz, n_passages, -1)
        scores = scores.view(n_layers, bsz, n_tokens, n_passages, -1)
        mask = mask.view(bsz, n_passages, -1)
        scores = scores.masked_fill(~mask[None, :, None], 0.0)

        scores = scores.sum(dim=[0])

        scores_woquery = None
        # Compute scores based on scores without query
        if not mask_query is None:
            output[f"{prefix}woquery"] = self.get_woquery_score(
                scores,
                mask_query,
                mask,
                n_layers,
                output_sequence_lengths=output_sequence_lengths,
            )

        return output

    def get_woquery_score(
        self, scores, mask_query, mask, n_layers, output_sequence_lengths
    ):
        if scores.size(-1) > mask_query.size(-1):
            zero_padding = torch.zeros(
                [mask_query.size(0), scores.size(-1) - mask_query.size(-1)],
                device=mask_query.device,
                dtype=torch.bool,
            )
            mask_query = torch.cat([mask_query, zero_padding], dim=-1)
        mask_query = mask * (~mask_query[:, None])
        scores_woquery = scores.masked_fill(~mask_query[:, None], 0.0)

        ntokens_woquery = 256 * n_layers

        # zero out scores after EOS token. This is needed when batching results in sequences with different lengths.
        for i in range(len(scores_woquery)):
            scores_woquery[i, output_sequence_lengths[i] :, :, :] = 0

        scores_woquery = scores_woquery.sum(dim=[1, 3])
        return scores_woquery / ntokens_woquery

    def overwrite_forward_crossattention(self):
        """
        Replace cross-attention forward function, only used to save
        cross-attention scores.
        """
        for mod in self.decoder.block:
            xattn = mod.layer[1].EncDecAttention
            xattn.forward = types.MethodType(cross_attention_forward, xattn)

    def create_crossattention_storage(self):
        for mod in self.decoder.block:
            xattn = mod.layer[1].EncDecAttention
            xattn.normalized_score_storage = None


def cross_attention_forward(
    self,
    hidden_states,
    mask=None,
    key_value_states=None,
    position_bias=None,
    past_key_value=None,
    layer_head_mask=None,
    query_length=None,
    use_cache=False,
    output_attentions=False,
    cache_position=None,
):
    """
    Self-attention (if key_value_states is None) or attention over source sentence (provided by key_value_states).
    """
    # Input is (batch_size, seq_length, dim)
    # Mask is (batch_size, key_length) (non-causal) or (batch_size, key_length, key_length)
    # past_key_value[0] is (batch_size, n_heads, q_len - 1, dim_per_head)

    batch_size, seq_length = hidden_states.shape[:2]
    real_seq_length = seq_length

    if past_key_value is not None:
        assert (
            len(past_key_value) == 2
        ), f"past_key_value should have 2 past states: keys and values. Got {len(past_key_value)} past states"
        real_seq_length += (
            past_key_value[0].shape[2] if query_length is None else query_length
        )

    key_length = (
        real_seq_length if key_value_states is None else key_value_states.shape[1]
    )

    def shape(states):
        """projection"""
        return states.view(
            batch_size, -1, self.n_heads, self.key_value_proj_dim
        ).transpose(1, 2)

    def unshape(states):
        """reshape"""
        return states.transpose(1, 2).contiguous().view(batch_size, -1, self.inner_dim)

    def project(hidden_states, proj_layer, key_value_states, past_key_value):
        """projects hidden states correctly to key/query states"""
        if key_value_states is None:
            # self-attn
            # (batch_size, n_heads, seq_length, dim_per_head)
            hidden_states = shape(proj_layer(hidden_states))
        elif past_key_value is None:
            # cross-attn
            # (batch_size, n_heads, seq_length, dim_per_head)
            hidden_states = shape(proj_layer(key_value_states))

        if past_key_value is not None:
            if key_value_states is None:
                # self-attn
                # (batch_size, n_heads, key_length, dim_per_head)
                hidden_states = torch.cat([past_key_value, hidden_states], dim=2)
            else:
                # cross-attn
                hidden_states = past_key_value
        return hidden_states

    # get query states
    query_states = shape(
        self.q(hidden_states)
    )  # (batch_size, n_heads, seq_length, dim_per_head)

    # get key/value states
    key_states = project(
        hidden_states,
        self.k,
        key_value_states,
        past_key_value[0] if past_key_value is not None else None,
    )
    value_states = project(
        hidden_states,
        self.v,
        key_value_states,
        past_key_value[1] if past_key_value is not None else None,
    )

    # compute scores
    scores = torch.matmul(
        query_states, key_states.transpose(3, 2)
    )  # equivalent of torch.einsum("bnqd,bnkd->bnqk", query_states, key_states), compatible with onnx op>9

    if position_bias is None:
        if not self.has_relative_attention_bias:
            position_bias = torch.zeros(
                (1, self.n_heads, real_seq_length, key_length),
                device=scores.device,
                dtype=scores.dtype,
            )
            if self.gradient_checkpointing and self.training:
                position_bias.requires_grad = True
        else:
            position_bias = self.compute_bias(real_seq_length, key_length)

        # if key and values are already calculated
        # we want only the last query position bias
        if past_key_value is not None:
            position_bias = position_bias[:, :, -hidden_states.size(1) :, :]

        if mask is not None:
            position_bias = (
                position_bias + mask
            )  # (batch_size, n_heads, seq_length, key_length)

    scores += position_bias

    attn_weights = nn.functional.softmax(scores.float(), dim=-1)  # .type_as(scores)

    if hasattr(self, "normalized_score_storage"):
        with torch.no_grad():
            self.normalized_score_storage = (
                (torch.norm(value_states.float(), dim=-1)[:, :, None] * attn_weights)
                .detach()
                .mean(dim=1)
            )

    attn_weights = nn.functional.dropout(
        attn_weights.type_as(scores), p=self.dropout, training=self.training
    )

    # Mask heads if we want to
    if layer_head_mask is not None:
        attn_weights = attn_weights * layer_head_mask

    attn_output = unshape(
        torch.matmul(attn_weights, value_states)
    )  # (batch_size, seq_length, dim)
    attn_output = self.o(attn_output)

    present_key_value_state = (
        (key_states, value_states) if (self.is_decoder and use_cache) else None
    )
    outputs = (attn_output,) + (present_key_value_state,) + (position_bias,)

    if output_attentions:
        outputs = outputs + (attn_weights,)
    return outputs



import json
from dataclasses import dataclass, field
from typing import Any, Dict, List, Union

from dacite import from_dict


@dataclass
class Query:
    text: str
    qid: Union[str | int]


@dataclass
class Candidate:
    docid: Union[str | int]
    score: float
    doc: Dict[str, Any]


@dataclass
class Request:
    query: Query
    candidates: List[Candidate] = field(default_factory=list)


@dataclass
class InferenceInvocation:
    prompt: Any
    response: str
    input_token_count: int
    output_token_count: int


@dataclass
class Result:
    query: Query
    candidates: list[Candidate] = field(default_factory=list)
    invocations_history: list[InferenceInvocation] = (field(default_factory=list),)


def read_requests_from_file(file_path: str) -> List[Request]:
    extension = file_path.split(".")[-1]
    if extension == "jsonl":
        requests = []
        with open(file_path, "r") as f:
            for l in f:
                if not l.strip():
                    continue
                requests.append(from_dict(data_class=Request, data=json.loads(l)))
        return requests
    elif extension == "json":
        with open(file_path, "r") as f:
            request_dicts = json.load(f)
        return [
            from_dict(data_class=Request, data=request_dict)
            for request_dict in request_dicts
        ]
    else:
        raise ValueError(f"Expected json or jsonl file format, got {extension}")


class DataWriter:
    def __init__(
        self,
        data: Union[Request | Result | List[Result] | List[Request]],
        append: bool = False,
    ):
        if isinstance(data, list):
            self._data = data
        else:
            self._data = [data]
        self._append = append

    def write_inference_invocations_history(self, filename: str):
        aggregated_history = []
        for d in self._data:
            values = []
            for info in d.invocations_history:
                values.append(info.__dict__)
            aggregated_history.append(
                {"query": d.query.__dict__, "invocations_history": values}
            )
        with open(filename, "a" if self._append else "w") as f:
            json.dump(aggregated_history, f, indent=2)

    def write_in_json_format(self, filename: str):
        results = []
        for d in self._data:
            candidates = [candidate.__dict__ for candidate in d.candidates]
            results.append({"query": d.query.__dict__, "candidates": candidates})
        with open(filename, "a" if self._append else "w") as f:
            json.dump(results, f, indent=2)

    def write_in_jsonl_format(self, filename: str):
        with open(filename, "a" if self._append else "w") as f:
            for d in self._data:
                candidates = [candidate.__dict__ for candidate in d.candidates]
                json.dump({"query": d.query.__dict__, "candidates": candidates}, f)
                f.write("\n")

    def write_in_trec_eval_format(self, filename: str):
        with open(filename, "a" if self._append else "w") as f:
            for d in self._data:
                qid = d.query.qid
                for rank, cand in enumerate(d.candidates, start=1):
                    f.write(f"{qid} Q0 {cand.docid} {rank} {cand.score} rank_llm\n")



from typing import Any, Dict, List, Optional, Tuple, Union

import torch
from tqdm import tqdm
from transformers import T5Tokenizer

from rank_llm.data import Request, Result
from rank_llm.rerank.listwise.listwise_rankllm import ListwiseRankLLM
from rank_llm.rerank.listwise.lit5.model import FiD, FiDCrossAttentionScore
from rank_llm.rerank.rankllm import PromptMode


class RankFiDDistill(ListwiseRankLLM):
    def _post_init(self):
        self._to_precision(self._precision)

    def _tokenize(self, s: str):
        return self._tokenizer(s)

    def _to_precision(self, precision: str) -> None:
        """
        We don't support python12 for now, after python 12, the code should be changed into
        """
        if precision == "float32":
            self._llm = self._llm.float()
        elif precision == "bfloat16":
            self._llm = self._llm.bfloat16()
        elif precision == "float16":
            self._llm = self._llm.float16()

    def __init__(
        self,
        model: str,
        context_size: int = 150,
        prompt_mode: PromptMode = PromptMode.LiT5,  # Placeholder for actual mode
        num_few_shot_examples: int = 0,
        window_size: int = 20,
        step_size: int = 10,
        precision: str = "bfloat16",
        device: str = "cuda",
        batched: bool = False,
    ) -> None:
        """
        Creates instance of the RankFiDDistill class, a specialized version of RankLLM designed from Lit5-Distill.
        """
        super().__init__(
            model=model,
            context_size=context_size,
            prompt_mode=prompt_mode,
            num_few_shot_examples=num_few_shot_examples,
            window_size=window_size,
        )
        self._precision = precision
        self._tokenizer = T5Tokenizer.from_pretrained(model)
        self._llm = FiD.from_pretrained(model).to(device).eval()

        self._device = device

        self._window_size = window_size
        self._stride = step_size

        self._batched = batched

        self._answer_maxlength = len(
            " > ".join(map(lambda x: f"[{x}]", range(1, window_size + 1)))
        )

        self._output_token_estimate = None

        self._post_init()

    def _run_llm_by_length_unified(
        self, batch_prompts: List[List[str]]
    ) -> List[Tuple[str, int]]:
        if len(batch_prompts) == 0:
            return []

        self._llm.eval()

        batch_size = len(batch_prompts)
        n_passages = len(batch_prompts[0])

        # single batch, unsqueeze
        inputs = {
            k: v.reshape(batch_size, -1).to(self._device)
            for k, v in self._tokenizer(
                [prompt for prompts in batch_prompts for prompt in prompts],
                return_tensors="pt",
                padding="max_length",
                truncation=True,
                max_length=self.max_tokens(),
            ).items()
        }

        with torch.no_grad():
            self._llm.reset_n_passages(n_passages=n_passages)
            outputs = self._llm.generate(
                **inputs,
                max_length=self._answer_maxlength,
                do_sample=False,
            )

        decoded_outputs = [
            self._tokenizer.decode(outputs[i], skip_special_tokens=True)
            for i in range(outputs.shape[0])
        ]

        # all token size should be equal
        return [
            (decoded_output, outputs.shape[1]) for decoded_output in decoded_outputs
        ]

    def rerank_batch(
        self,
        requests: List[Request],
        rank_start: int = 0,
        rank_end: int = 100,
        shuffle_candidates: bool = False,
        logging: bool = False,
        **kwargs: Any,
    ) -> List[Result]:
        top_k_retrieve: int = kwargs.get("top_k_retrieve", rank_end)
        rank_end = min(top_k_retrieve, rank_end)
        window_size: int = kwargs.get("window_size", self._window_size)
        window_size = min(window_size, top_k_retrieve)
        step: int = kwargs.get("step_size", self._stride)
        populate_invocations_history: bool = kwargs.get(
            "populate_invocations_history", False
        )
        batch_size = kwargs.get("batch_size", 1)

        if self._batched:
            # reranking using vllm
            if len(set([len(req.candidates) for req in requests])) != 1:
                raise ValueError(
                    "Batched requests must have the same number of candidates"
                )

            result = []

            with tqdm(range(0, len(requests))) as bar:
                for i in range(0, len(requests), batch_size):
                    batch = requests[i : min(i + batch_size, len(requests))]
                    batch_result = self.sliding_windows_batched(
                        batch,
                        rank_start=max(rank_start, 0),
                        rank_end=min(
                            rank_end, len(requests[0].candidates)
                        ),  # TODO: Fails arbitrary hit sizes
                        window_size=window_size,
                        step=step,
                        shuffle_candidates=shuffle_candidates,
                        logging=logging,
                        populate_invocations_history=populate_invocations_history,
                    )
                    result.extend(batch_result)
                    bar.update(len(batch))

            return result
        else:
            # Normal operation mode
            results = []
            for request in tqdm(requests):
                result = self.sliding_windows(
                    request,
                    rank_start=max(rank_start, 0),
                    rank_end=min(rank_end, len(request.candidates)),
                    window_size=window_size,
                    step=step,
                    shuffle_candidates=shuffle_candidates,
                    logging=logging,
                    populate_invocations_history=populate_invocations_history,
                )
                results.append(result)
            return results

    def run_llm_batched(
        self, prompts: List[List[Dict[str, str]]], **kwargs
    ) -> List[Tuple[str, int]]:
        if len(prompts) == 0:
            return []

        # unfortunately, we are not allowed to use VLLM on T5. However, we could unify the prompts by passage size
        #   (which is commonly the same) then rerank stuff having same passage sizes

        prompt_infos = [list(map(lambda x: x["text"], prompt)) for prompt in prompts]

        return self._run_llm_by_length_unified(prompt_infos)

    def create_prompt_batched(
        self, results: List[Result], rank_start: int, rank_end: int, batch_size: int
    ) -> List[Tuple[List[Dict[str, str]], int]]:
        return [self.create_prompt(result, rank_start, rank_end) for result in results]

    def run_llm(self, prompts: List[Dict[str, str]], **kwargs) -> Tuple[str, int]:
        """
        Run the target language model with a passed in prompt.
        """

        return self._run_llm_by_length_unified(
            [list(map(lambda x: x["text"], prompts))]
        )[0]

    def create_prompt(
        self, result: Result, rank_start: int, rank_end: int
    ) -> Tuple[List[Dict[str, str]], int]:
        """
        Create a prompt based on the result and given ranking range.
        """

        # For now, we concat the prompt, because it seems LiT5 is also concatting the stuff
        prompts = [
            {
                "text": self._gen_passage(
                    result.query.text,
                    i + 1 - rank_start,
                    self.convert_doc_to_prompt_content(
                        result.candidates[i].doc, self.max_tokens()
                    ),
                )
            }
            for i in range(rank_start, rank_end)
        ]

        return prompts, sum(self.get_num_tokens(prompt["text"]) for prompt in prompts)

    def get_num_tokens(self, prompt: Union[str, List[Dict[str, str]]]) -> int:
        """
        Abstract method to calculate the number of tokens contained in the given prompt.
        """
        if isinstance(prompt, str):
            return len(self._tokenizer.encode(prompt))
        elif isinstance(prompt, list):
            return sum(len(self._tokenizer.encode(item["text"])) for item in prompt)
        else:
            raise ValueError(
                "Prompt must be a string or a list of dictionaries with a 'text' key."
            )

    def cost_per_1k_token(self, input_token: bool) -> float:
        return 0

    def num_output_tokens(self, current_window_size: Optional[int] = None) -> int:
        if current_window_size is None:
            current_window_size = self._window_size
        if (
            self._output_token_estimate is not None
            and self._window_size == current_window_size
        ):
            return self._output_token_estimate
        else:
            output_token_estimate = (
                len(
                    self._tokenizer.encode(
                        " > ".join([f"[{i + 1}]" for i in range(current_window_size)])
                    )
                )
                - 1
            )
            if (
                self._output_token_estimate is None
                and self._window_size == current_window_size
            ):
                self._output_token_estimate = output_token_estimate

            return output_token_estimate

    @staticmethod
    def _gen_passage(query: str, index: int, passage: str) -> str:
        return f"Search Query: {query} Passage: [{index}] {passage} Relevance Ranking: "


class RankFiDScore(ListwiseRankLLM):
    def _post_init(self):
        # set the overwrite forward cross attention
        self._llm.overwrite_forward_crossattention()
        self._to_precision(self._precision)

    def _tokenize(self, s: str):
        return self._tokenizer(s)

    def _to_precision(self, precision: str) -> None:
        """
        We don't support python12 for now, after python 12, the code should be changed into
        """
        if precision == "float32":
            self._llm = self._llm.float()
        elif precision == "bfloat16":
            self._llm = self._llm.bfloat16()
        elif precision == "float16":
            self._llm = self._llm.float16()

    def __init__(
        self,
        model: str,
        context_size: int = 150,
        prompt_mode: PromptMode = PromptMode.LiT5,  # Placeholder for actual mode
        num_few_shot_examples: int = 0,
        window_size: int = 20,
        step_size: int = 10,
        precision: str = "bfloat16",
        device: str = "cuda",
        batched: bool = False,
    ) -> None:
        super().__init__(
            model=model,
            context_size=context_size,
            prompt_mode=prompt_mode,
            num_few_shot_examples=num_few_shot_examples,
            window_size=window_size,
        )
        self._precision = precision
        self._tokenizer = T5Tokenizer.from_pretrained(model)
        self._llm = FiDCrossAttentionScore.from_pretrained(model).to(device).eval()

        self._device = device
        self._window_size = window_size
        self._stride = step_size

        self._batched = batched

        self._output_token_estimate = None

        self._post_init()

    def _run_llm_by_length_unified(
        self, batch_prompts: List[List[Tuple[str, str]]]
    ) -> List[Tuple[str, int]]:
        if len(batch_prompts) == 0:
            return []

        # get arbitrary query (they should be the same)
        queries = [prompts[0][0] for prompts in batch_prompts]
        batch_size = len(batch_prompts)
        n_passages = len(batch_prompts[0])

        inputs = {
            k: v.reshape(batch_size, -1).to(self._device)
            for k, v in self._tokenizer(
                [prompt for prompts in batch_prompts for (_, prompt) in prompts],
                return_tensors="pt",
                padding="max_length",
                truncation=True,
                max_length=self.max_tokens(),
            ).items()
        }

        passage_ids = inputs["input_ids"]
        passage_mask = inputs["attention_mask"]

        with torch.no_grad():
            self._llm.reset_score_storage()

            outputs = self._llm.generate(
                **inputs, max_length=20, do_sample=False, n_passages=n_passages
            )

        output_sequence_lengths = []

        for output in outputs:
            output_length = 0
            for j in range(output.shape[0]):
                if output[j] == FiDCrossAttentionScore.ANSWER_EOS_TOKEN:
                    output_length = j
                    break
            else:
                output_length = outputs.shape[1]
            output_sequence_lengths.append(output_length)

        query_mask_reader = self._tokenizer(
            queries,
            max_length=self.max_tokens(),
            padding="longest",
            truncation=True,
            return_tensors="pt",
            add_special_tokens=False,
        )["attention_mask"].bool()

        with torch.no_grad():
            crossattention_scores = self._llm.get_crossattention_scores(
                n_passages,
                ids=passage_ids.to(self._device),
                mask=passage_mask.bool().to(self._device),
                mask_query=query_mask_reader.to(self._device),
                output_sequence_lengths=output_sequence_lengths,
            )
            # only supports normswoquery for now
            crossattention_score: torch.Tensor = crossattention_scores["normswoquery"]
            sorted, idxes = torch.sort(crossattention_score, dim=-1, descending=True)
            idxes = idxes.detach().cpu()

        return [
            (
                " > ".join([f"[{x + 1}]" for x in idxes[i].tolist()]),
                output_sequence_lengths[i] + crossattention_score.shape[1],
            )
            for i in range(idxes.shape[0])
        ]

    def rerank_batch(
        self,
        requests: List[Request],
        rank_start: int = 0,
        rank_end: int = 100,
        shuffle_candidates: bool = False,
        logging: bool = False,
        **kwargs: Any,
    ) -> List[Result]:
        top_k_retrieve: int = kwargs.get("top_k_retrieve", rank_end)
        rank_end = min(top_k_retrieve, rank_end)
        window_size: int = kwargs.get("window_size", self._window_size)
        window_size = min(window_size, top_k_retrieve)
        step: int = kwargs.get("step_size", self._stride)
        populate_invocations_history: bool = kwargs.get(
            "populate_invocations_history", False
        )
        batch_size = kwargs.get("batch_size", 1)

        if self._batched:
            # reranking using vllm
            if len(set([len(req.candidates) for req in requests])) != 1:
                raise ValueError(
                    "Batched requests must have the same number of candidates"
                )

            result = []

            with tqdm(range(0, len(requests))) as bar:
                for i in range(0, len(requests), batch_size):
                    batch = requests[i : min(i + batch_size, len(requests))]
                    batch_result = self.sliding_windows_batched(
                        batch,
                        rank_start=max(rank_start, 0),
                        rank_end=min(
                            rank_end, len(requests[0].candidates)
                        ),  # TODO: Fails arbitrary hit sizes
                        window_size=window_size,
                        step=step,
                        shuffle_candidates=shuffle_candidates,
                        logging=logging,
                        populate_invocations_history=populate_invocations_history,
                    )
                    result.extend(batch_result)
                    bar.update(len(batch))

            return result
        else:
            # Normal operation mode
            results = []
            for request in tqdm(requests):
                result = self.sliding_windows(
                    request,
                    rank_start=max(rank_start, 0),
                    rank_end=min(rank_end, len(request.candidates)),
                    window_size=window_size,
                    step=step,
                    shuffle_candidates=shuffle_candidates,
                    logging=logging,
                    populate_invocations_history=populate_invocations_history,
                )
                results.append(result)
            return results

    def run_llm_batched(
        self, prompts: List[List[Dict[str, str]]], **kwargs
    ) -> List[Tuple[str, int]]:
        if len(prompts) == 0:
            return []

        # unfortunately, we are not allowed to use VLLM on T5. However, we could unify the prompts by passage size
        #   (which is commonly the same) then rerank stuff having same passage sizes

        processed_prompts = [
            [(x["query"], x["text"]) for x in prmpt] for prmpt in prompts
        ]

        return self._run_llm_by_length_unified(processed_prompts)

    def create_prompt_batched(
        self, results: List[Result], rank_start: int, rank_end: int, batch_size: int
    ) -> List[Tuple[List[Dict[str, str]], int]]:
        return [self.create_prompt(result, rank_start, rank_end) for result in results]

    def run_llm(self, prompts: List[Dict[str, str]], **kwargs) -> Tuple[str, int]:
        # get arbitrary query (they should be the same)
        return self._run_llm_by_length_unified(
            [[(x["query"], x["text"]) for x in prompts]]
        )[0]

    def create_prompt(
        self, result: Result, rank_start: int, rank_end: int
    ) -> Tuple[List[Dict[str, str]], int]:
        """
        Create a prompt based on the result and given ranking range.
        """
        query = result.query.text
        results = []

        sum_token = 0

        for i in range(rank_start, rank_end):
            results.append(
                {
                    "query": f"question: {query}",
                    "text": self._gen_passage(
                        query,
                        self.convert_doc_to_prompt_content(
                            result.candidates[i].doc, self.max_tokens()
                        ),
                    ),
                }
            )
            sum_token += len(self._tokenizer.encode(results[-1]["text"]))

        return results, sum_token

    def get_num_tokens(self, prompt: str) -> int:
        return len(self._tokenizer.encode(prompt))

    def cost_per_1k_token(self, input_token: bool) -> float:
        return 0.0

    def num_output_tokens(self, current_window_size: Optional[int] = None) -> int:
        if current_window_size is None:
            current_window_size = self._window_size
        if (
            self._output_token_estimate is not None
            and self._window_size == current_window_size
        ):
            return self._output_token_estimate
        else:
            output_token_estimate = (
                len(
                    self._tokenizer.encode(
                        " > ".join([f"[{i + 1}]" for i in range(current_window_size)])
                    )
                )
                - 1
            )
            if (
                self._output_token_estimate is None
                and self._window_size == current_window_size
            ):
                self._output_token_estimate = output_token_estimate

            return output_token_estimate

    @staticmethod
    def _gen_passage(query: str, passage: str) -> str:
        return f"question: {query} context: {passage}"



import logging
from abc import ABC, abstractmethod
from enum import Enum
from typing import Any, Dict, List, Tuple, Union

from rank_llm.data import Request, Result

logger = logging.getLogger(__name__)


class PromptMode(Enum):
    UNSPECIFIED = "unspecified"
    RANK_GPT = "rank_GPT"
    RANK_GPT_APEER = "rank_GPT_APEER"
    LRL = "LRL"
    MONOT5 = "monot5"
    DUOT5 = "duot5"
    LiT5 = "LiT5"

    def __str__(self):
        return self.value


class RankLLM(ABC):
    def __init__(self, model: str, context_size: int, prompt_mode: PromptMode) -> None:
        self._model = model
        self._context_size = context_size
        self._prompt_mode = prompt_mode

    @abstractmethod
    def run_llm_batched(
        self, prompts: List[Union[str, List[Dict[str, str]]]], **kwargs
    ) -> List[Tuple[str, int]]:
        """
        Abstract method to run the target language model with a batch of prompts.

        Args:
            prompts (List[Union[str, List[Dict[str, str]]]): The list of prompts to be processed by the model.

        Returns:
            List[Tuple[str, int]]: A list of tuple objects containing the text responses and the number of tokens in the responses.
        """
        pass

    @abstractmethod
    def run_llm(
        self, prompt: Union[str, List[Dict[str, str]]], **kwargs
    ) -> Tuple[str, int]:
        """
        Abstract method to run the target language model with a passed in prompt.

        Args:
            prompt (Union[str, List[Dict[str, str]]]): The prompt to be processed by the model.

        Returns:
            Tuple[str, int]: A tuple object containing the text response and the number of tokens in the response.
        """
        pass

    @abstractmethod
    def create_prompt_batched(
        self, results: List[Result], rank_start: int, rank_end: int, batch_size: int
    ) -> List[Tuple[Union[str, List[Dict[str, str]]], int]]:
        """
        Abstract method to create a batch of prompts based on the results and given ranking range.

        Args:
            results (List[Result]): The list of result objects containing data for prompt generation.
            rank_start (int): The starting rank for prompt generation.
            rank_end (int): The ending rank for prompt generation.

        Returns:
            Tuple[List[Union[str, List[Dict[str, str]]], List[int]]: A tuple object containing the list of generated prompts and the list of number of tokens in the generated prompts.
        """
        pass

    @abstractmethod
    def create_prompt(
        self, result: Result, rank_start: int, rank_end: int
    ) -> Tuple[Union[str, List[Dict[str, str]]], int]:
        """
        Abstract method to create a prompt based on the result and given ranking range.

        Args:
            result (Result): The result object containing data for prompt generation.
            rank_start (int): The starting rank for prompt generation.
            rank_end (int): The ending rank for prompt generation.

        Returns:
            Tuple[Union[str, List[Dict[str, str]]], int]: A tuple object containing the generated prompt and the number of tokens in the generated prompt.
        """
        pass

    @abstractmethod
    def get_num_tokens(self, prompt: Union[str, List[Dict[str, str]]]) -> int:
        """
        Abstract method to calculate the number of tokens contained in the given prompt.

        Args:
            prompt (Union[str, List[Dict[str, str]]]): The prompt for which to compute the token count for.

        Returns:
            int: The number of tokens in the given prompt.
        """
        pass

    @abstractmethod
    def cost_per_1k_token(self, input_token: bool) -> float:
        """
        Abstract method to calculate the cost per 1,000 tokens for the target language model.

        Args:
            input_token (bool): Flag to indicate if the cost is for input tokens or output tokens.

        Returns:
            float: The cost per 1,000 tokens.
        """
        pass

    @abstractmethod
    def num_output_tokens(self) -> int:
        """
        Abstract method to estimate the number of tokens in the model's output, constrained by max tokens for the target language model.

        Returns:
            int: The estimated number of output tokens.
        """
        pass

    @abstractmethod
    def rerank_batch(
        self,
        requests: List[Request],
        rank_start: int = 0,
        rank_end: int = 100,
        shuffle_candidates: bool = False,
        logging: bool = False,
        **kwargs: Any,
    ) -> List[Result]:
        """
        Reranks a list of requests using the RankLLM model_coordinator.

        This function applies a sliding window algorithm to rerank the results.
        Each window of results is processed by the RankLLM model_coordinator to obtain a new ranking.

        Args:
            requests (List[Request]): The list of requests. Each request has a query and a candidates list.
            rank_start (int, optional): The starting rank for processing. Defaults to 0.
            rank_end (int, optional): The end rank for processing. Defaults to 100.
            window_size (int, optional): The size of each sliding window. Defaults to 20.
            step (int, optional): The step size for moving the window. Defaults to 10.
            shuffle_candidates (bool, optional): Whether to shuffle candidates before reranking. Defaults to False.
            logging (bool, optional): Enables logging of the reranking process. Defaults to False.
            vllm_batched (bool, optional): Whether to use VLLM batched processing. Defaults to False.
            sglang_batched (bool, optional): Whether to use SGLang batched processing. Defaults to False.
            tensorrt_batched (bool, optional): Whether to use TensorRT-LLM batches processing. Defaults to False.
            populate_invocations_history (bool, optional): Whether to populate the history of inference invocations. Defaults to False.
            batched (bool, optional): Whether to use batched processing. Defaults to False.

        Returns:
            List[Result]: A list containing the reranked candidates.
        """
        pass

    @abstractmethod
    def get_output_filename(
        self,
        top_k_candidates: int,
        dataset_name: str,
        shuffle_candidates: bool,
        **kwargs: Any,
    ) -> str:
        """
        Returns the output filename used when writing rerank results to file
        """
        pass



import json
import logging
import os
import random
import unicodedata
from concurrent.futures import ThreadPoolExecutor
from typing import Any, Dict, List, Optional, Tuple

import torch
from fastchat.model import get_conversation_template, load_model
from ftfy import fix_text
from tqdm import tqdm
from transformers.generation import GenerationConfig

from rank_llm.data import Request, Result
from rank_llm.rerank import PromptMode

from .listwise_rankllm import ListwiseRankLLM

try:
    from vllm import LLM, RequestOutput, SamplingParams
except:
    LLM = None
    RequestOutput = None
    SamplingParams = None

logger = logging.getLogger(__name__)

ALPH_START_IDX = ord("A") - 1


class RankListwiseOSLLM(ListwiseRankLLM):
    def __init__(
        self,
        model: str,
        name: str = "",
        context_size: int = 4096,
        prompt_mode: PromptMode = PromptMode.RANK_GPT,
        num_few_shot_examples: int = 0,
        device: str = "cuda",
        num_gpus: int = 1,
        variable_passages: bool = False,
        window_size: int = 20,
        system_message: str = None,
        use_logits: bool = False,
        use_alpha: bool = False,
        vllm_batched: bool = False,
        sglang_batched: bool = False,
        tensorrt_batched: bool = False,
    ) -> None:
        """
         Creates instance of the RankListwiseOSLLM class, an extension of RankLLM designed for performing listwise ranking of passages using a specified language model. Advanced configurations are supported such as GPU acceleration, variable passage handling, and custom system messages for generating prompts.
         RankListWiseOSLLM uses the default implementations for sliding_window

         Parameters:
         - model (str): Identifier for the language model to be used for ranking tasks.
         - context_size (int, optional): Maximum number of tokens that can be handled in a single prompt. Defaults to 4096.
        - prompt_mode (PromptMode, optional): Specifies the mode of prompt generation, with the default set to RANK_GPT,
         indicating that this class is designed primarily for listwise ranking tasks following the RANK_GPT methodology.
         - num_few_shot_examples (int, optional): Number of few-shot learning examples to include in the prompt, allowing for
         the integration of example-based learning to improve model performance. Defaults to 0, indicating no few-shot examples
         by default.
         - device (str, optional): Specifies the device for model computation ('cuda' for GPU or 'cpu'). Defaults to 'cuda'.
         - num_gpus (int, optional): Number of GPUs to use for model loading and inference. Defaults to 1.
         - variable_passages (bool, optional): Indicates whether the number of passages to rank can vary. Defaults to False.
         - window_size (int, optional): The window size for handling text inputs. Defaults to 20.
         - system_message (Optional[str], optional): Custom system message to be included in the prompt for additional
         instructions or context. Defaults to None.
         - use_logits (bool, optional): Indicates whether to use logits or not. Defaults to False.
         - use_alpha (bool, optional): Indicates whether to use alphabet ordering the prompts. Defaults to False.
         - vllm_batched (bool, optional): Indicates whether batched inference using VLLM is leveraged. Defaults to False.
         - sglang_batched (bool, optional): Indicates whether batched inference using SGLang is leveraged. Defaults to False.
         - tensorrt_batched (bool, optional): Indicates whether batched inference using TensorRT-LLM is leveraged. Defaults to False.

         Raises:
         - AssertionError: If CUDA is specified as the device but is not available on the system.
         - ValueError: If an unsupported prompt mode is provided.

         Note:
         - This class is operates given scenarios where listwise ranking is required, with support for dynamic
         passage handling and customization of prompts through system messages and few-shot examples.
         - GPU acceleration is supported and recommended for faster computations.
        TODO: Make repetition_penalty configurable
        """
        super().__init__(
            model,
            context_size,
            prompt_mode,
            num_few_shot_examples,
            window_size,
            use_alpha=use_alpha,
        )
        self._device = device
        self._vllm_batched = vllm_batched
        self._sglang_batched = sglang_batched
        self._tensorrt_batched = tensorrt_batched
        self._name = name
        self._variable_passages = variable_passages
        self._system_message = system_message
        self._output_token_estimate = None
        self._use_logits = use_logits

        if num_few_shot_examples > 0:
            with open("data/output_v2_aug_filtered.jsonl", "r") as json_file:
                self._examples = list(json_file)[1:-1]
        if self._device == "cuda":
            assert torch.cuda.is_available()

        if prompt_mode != PromptMode.RANK_GPT:
            raise ValueError(
                f"Unsupported prompt mode: {prompt_mode}. The only prompt mode currently supported is a slight variation of {PromptMode.RANK_GPT} prompt."
            )
        if vllm_batched and LLM is None:
            raise ImportError(
                "Please install rank-llm with `pip install -e .[vllm]` to use batch inference."
            )
        elif vllm_batched:
            self._llm = LLM(
                model,
                download_dir=os.getenv("HF_HOME"),
                enforce_eager=False,
                max_logprobs=30,
                tensor_parallel_size=num_gpus,
            )
            self._tokenizer = self._llm.get_tokenizer()
        elif sglang_batched:
            try:
                from sglang import Engine
            except Exception:
                raise ImportError(
                    "Please install rank-llm with `pip install -e .[sglang]` to use sglang batch inference."
                )
            port = random.randint(30000, 35000)
            self._llm = Engine(
                model_path=model, port=port, download_dir=os.getenv("HF_HOME")
            )
            self._tokenizer = self._llm.get_tokenizer()
        elif tensorrt_batched:
            try:
                from tensorrt_llm import LLM as TRTLLM
                from tensorrt_llm import BuildConfig
            except Exception:
                raise ImportError(
                    "Please install rank-llm with `pip install -e .[tensorrt-llm]` to use tensorrt batch inference."
                )
            build_config = BuildConfig(max_seq_len=4096)
            self._llm = TRTLLM(model=model, build_config=build_config)
            self._tokenizer = self._llm.tokenizer
        else:
            self._llm, self._tokenizer = load_model(
                model, device=device, num_gpus=num_gpus
            )

    def rerank_batch(
        self,
        requests: List[Request],
        rank_start: int = 0,
        rank_end: int = 100,
        shuffle_candidates: bool = False,
        logging: bool = False,
        **kwargs: Any,
    ) -> List[Result]:
        top_k_retrieve: int = kwargs.get("top_k_retrieve", rank_end)
        rank_end = min(top_k_retrieve, rank_end)
        window_size: int = kwargs.get("window_size", 20)
        window_size = min(window_size, top_k_retrieve)
        step: int = kwargs.get("step", 10)
        populate_invocations_history: bool = kwargs.get(
            "populate_invocations_history", False
        )
        if self._vllm_batched or self._sglang_batched or self._tensorrt_batched:
            # reranking using vllm or sglang or tensorrtllm
            if len(set([len(req.candidates) for req in requests])) != 1:
                raise ValueError(
                    "Batched requests must have the same number of candidates"
                )

            return self.sliding_windows_batched(
                requests,
                rank_start=max(rank_start, 0),
                rank_end=min(
                    rank_end, len(requests[0].candidates)
                ),  # TODO: Fails arbitrary hit sizes
                window_size=window_size,
                step=step,
                shuffle_candidates=shuffle_candidates,
                logging=logging,
                populate_invocations_history=populate_invocations_history,
            )
        else:
            if self._use_logits:
                raise TypeError(
                    "Reranking using logits of first identifier is currently only supported when vllm_batch=True"
                )

            # Normal operation mode
            results = []
            for request in tqdm(requests):
                result = self.sliding_windows(
                    request,
                    rank_start=max(rank_start, 0),
                    rank_end=min(rank_end, len(request.candidates)),
                    window_size=window_size,
                    step=step,
                    shuffle_candidates=shuffle_candidates,
                    logging=logging,
                    populate_invocations_history=populate_invocations_history,
                )
                results.append(result)
            return results

    def _evaluate_logits(
        self, logits: Dict[str, "Logit"], total: Tuple[int, int]
    ) -> Tuple[str, Dict[int, float]]:
        if self._use_alpha:
            evaluations = {
                ord(logit.decoded_token): logit.logprob
                for logit in logits.values()
                if len(logit.decoded_token) == 1
                and logit.decoded_token.isalpha()
                and ALPH_START_IDX + 1
                <= ord(logit.decoded_token)
                <= ALPH_START_IDX + self._window_size
            }
            sorted_evaluations = sorted(evaluations.items(), key=lambda x: -x[1])
            result_string = ">".join([f"[{chr(x)}]" for x, y in sorted_evaluations])
        else:
            evaluations = {
                int(logit.decoded_token): logit.logprob
                for logit in logits.values()
                if logit.decoded_token.isnumeric()
                and not unicodedata.name(logit.decoded_token).startswith(
                    ("SUPERSCRIPT", "VULGAR FRACTION", "SUBSCRIPT")
                )
                and total[0] <= int(logit.decoded_token) <= total[1]
            }
            sorted_evaluations = sorted(evaluations.items(), key=lambda x: -x[1])
            result_string = ">".join([f"[{x}]" for x, y in sorted_evaluations])

        return result_string, evaluations

    def _get_logits_single_digit(
        self,
        output: RequestOutput,
        effective_location: int = 1,
        total: Tuple[int, int] = (1, 9),
    ):
        logits = output.outputs[0].logprobs[effective_location]
        return self._evaluate_logits(logits, total)

    def run_llm_batched(
        self,
        prompts: List[str | List[Dict[str, str]]],
        current_window_size: Optional[int] = None,
    ) -> List[Tuple[str, int]]:
        if isinstance(self._llm, LLM) and self._vllm_batched:
            logger.info(f"VLLM Generating!")
            if current_window_size is None:
                current_window_size = self._window_size

            if self._use_logits:
                params = SamplingParams(
                    min_tokens=2, max_tokens=2, temperature=0.0, logprobs=30
                )
                outputs = self._llm.generate(prompts, sampling_params=params)
                arr = [self._get_logits_single_digit(output) for output in outputs]
                return [(s, len(s)) for s, __ in arr]
            else:
                sampling_params = SamplingParams(
                    temperature=0.0,
                    max_tokens=self.num_output_tokens(current_window_size),
                    min_tokens=self.num_output_tokens(current_window_size),
                )
                outputs = self._llm.generate(prompts, sampling_params)
                return [
                    (output.outputs[0].text, len(output.outputs[0].token_ids))
                    for output in outputs
                ]
        elif self._tensorrt_batched:
            import tensorrt_llm.hlapi.llm
            from tensorrt_llm import SamplingParams as TRTSamplingParams

            if isinstance(self._llm, tensorrt_llm.hlapi.llm.LLM):
                logger.info(f"TensorRT LLM Generating!")
                sampling_params = TRTSamplingParams(
                    temperature=0.0,
                    max_tokens=self.num_output_tokens(current_window_size),
                    min_tokens=self.num_output_tokens(current_window_size),
                )
                outputs = self._llm.generate(prompts, sampling_params)
                return [
                    (output.outputs[0].text, len(output.outputs[0].token_ids))
                    for output in outputs
                ]
        else:
            logger.info(f"SGLang Generating!")
            sampling_params = {
                "temperature": 0.0,
                "max_new_tokens": self.num_output_tokens(current_window_size),
                "min_new_tokens": self.num_output_tokens(current_window_size),
            }
            outputs = self._llm.generate(prompts, sampling_params)
            return [
                # completion_tokens counts stop token
                (output["text"], output["meta_info"]["completion_tokens"] - 1)
                for output in outputs
            ]

    def run_llm(
        self, prompt: str, current_window_size: Optional[int] = None
    ) -> Tuple[str, int]:
        if current_window_size is None:
            current_window_size = self._window_size

        if self._use_logits:
            params = SamplingParams(
                min_tokens=1, max_tokens=1, temperature=0.0, logprobs=30
            )
            output = self._llm.generate(
                [prompt + "["], sampling_params=params, use_tqdm=False
            )[0]
            s, _ = self._get_logits_single_digit(output, effective_location=0)
            return s, len(s)
        else:
            inputs = self._tokenizer([prompt])
            inputs = {k: torch.tensor(v).to(self._device) for k, v in inputs.items()}
            gen_cfg = GenerationConfig.from_model_config(self._llm.config)
            gen_cfg.max_new_tokens = self.num_output_tokens(current_window_size)
            gen_cfg.min_new_tokens = self.num_output_tokens(current_window_size)
            # gen_cfg.temperature = 0
            gen_cfg.do_sample = False
            output_ids = self._llm.generate(**inputs, generation_config=gen_cfg)

            if self._llm.config.is_encoder_decoder:
                output_ids = output_ids[0]
            else:
                output_ids = output_ids[0][len(inputs["input_ids"][0]) :]
            outputs = self._tokenizer.decode(
                output_ids,
                skip_special_tokens=True,
                spaces_between_special_tokens=False,
            )
            return outputs, output_ids.size(0)

    def num_output_tokens(self, current_window_size: Optional[int] = None) -> int:
        if current_window_size is None:
            current_window_size = self._window_size

        if self._output_token_estimate and self._window_size == current_window_size:
            return self._output_token_estimate

        if self._use_alpha:
            token_str = " > ".join(
                [f"[{chr(ALPH_START_IDX+i+1)}]" for i in range(current_window_size)]
            )
        else:
            token_str = " > ".join([f"[{i+1}]" for i in range(current_window_size)])

        _output_token_estimate = len(self._tokenizer.encode(token_str)) - 1

        if (
            self._output_token_estimate is None
            and self._window_size == current_window_size
        ):
            self._output_token_estimate = _output_token_estimate

        return _output_token_estimate

    def _add_prefix_prompt(self, query: str, num: int) -> str:
        identifier_type = "an alphabetical" if self._use_alpha else " a numerical"
        return f"I will provide you with {num} passages, each indicated by {identifier_type} identifier []. Rank the passages based on their relevance to the search query: {query}.\n"

    def _add_post_prompt(self, query: str, num: int) -> str:
        if self._use_alpha:
            example_ordering = "[B] > [A]" if self._variable_passages else "[D] > [B]"
        else:
            example_ordering = "[2] > [1]" if self._variable_passages else "[4] > [2]"
        return f"Search Query: {query}.\nRank the {num} passages above based on their relevance to the search query. All the passages should be included and listed using identifiers, in descending order of relevance. The output format should be [] > [], e.g., {example_ordering}, Only respond with the ranking results, do not say any word or explain."

    def _add_few_shot_examples(self, conv):
        for _ in range(self._num_few_shot_examples):
            ex = random.choice(self._examples)
            obj = json.loads(ex)
            prompt = obj["conversations"][0]["value"]
            response = obj["conversations"][1]["value"]
            conv.append_message(conv.roles[0], prompt)
            conv.append_message(conv.roles[1], response)
        return conv

    def _add_few_shot_examples_messages(self, messages):
        for _ in range(self._num_few_shot_examples):
            ex = random.choice(self._examples)
            obj = json.loads(ex)
            prompt = obj["conversations"][0]["value"]
            response = obj["conversations"][1]["value"]
            messages.append({"role": "user", "content": prompt})
            messages.append({"role": "assistant", "content": response})
        return messages

    def create_prompt(
        self, result: Result, rank_start: int, rank_end: int
    ) -> Tuple[str, int]:
        query = result.query.text
        query = self._replace_number(query)
        num = len(result.candidates[rank_start:rank_end])
        max_length = 300 * (20 / (rank_end - rank_start))
        while True:
            if self._vllm_batched:
                messages = list()
                if self._system_message:
                    messages.append({"role": "system", "content": self._system_message})
                messages = self._add_few_shot_examples_messages(messages)
                prefix = self._add_prefix_prompt(query, num)
                rank = 0
                input_context = f"{prefix}\n"
                for cand in result.candidates[rank_start:rank_end]:
                    rank += 1
                    content = self.convert_doc_to_prompt_content(cand.doc, max_length)

                    identifier = (
                        chr(ALPH_START_IDX + rank) if self._use_alpha else str(rank)
                    )
                    input_context += f"[{identifier}] {self._replace_number(content)}\n"

                input_context += self._add_post_prompt(query, num)
                messages.append({"role": "user", "content": input_context})

                prompt = self._tokenizer.apply_chat_template(
                    messages, tokenize=False, add_generation_prompt=True
                )
                prompt = fix_text(prompt)
                num_tokens = self.get_num_tokens(prompt)
                if num_tokens <= self.max_tokens() - self.num_output_tokens(
                    rank_end - rank_start
                ):
                    break
                else:
                    max_length -= max(
                        1,
                        (
                            num_tokens
                            - self.max_tokens()
                            + self.num_output_tokens(rank_end - rank_start)
                        )
                        // ((rank_end - rank_start) * 4),
                    )
            else:
                conv = get_conversation_template(self._model)
                if self._system_message:
                    conv.set_system_message(self._system_message)
                conv = self._add_few_shot_examples(conv)
                prefix = self._add_prefix_prompt(query, num)
                rank = 0
                input_context = f"{prefix}\n"
                for cand in result.candidates[rank_start:rank_end]:
                    rank += 1
                    # For Japanese should cut by character: content = content[:int(max_length)]
                    content = self.convert_doc_to_prompt_content(cand.doc, max_length)

                    identifier = (
                        chr(ALPH_START_IDX + rank) if self._use_alpha else str(rank)
                    )
                    input_context += f"[{identifier}] {self._replace_number(content)}\n"

                input_context += self._add_post_prompt(query, num)
                conv.append_message(conv.roles[0], input_context)
                conv.append_message(conv.roles[1], None)
                prompt = conv.get_prompt()
                prompt = fix_text(prompt)
                num_tokens = self.get_num_tokens(prompt)
                if num_tokens <= self.max_tokens() - self.num_output_tokens(
                    rank_end - rank_start
                ):
                    break
                else:
                    max_length -= max(
                        1,
                        (
                            num_tokens
                            - self.max_tokens()
                            + self.num_output_tokens(rank_end - rank_start)
                        )
                        // ((rank_end - rank_start) * 4),
                    )
        return prompt, self.get_num_tokens(prompt)

    def create_prompt_batched(
        self,
        results: List[Result],
        rank_start: int,
        rank_end: int,
        batch_size: int = 32,
    ) -> List[Tuple[str, int]]:
        def chunks(lst, n):
            """Yield successive n-sized chunks from lst."""
            for i in range(0, len(lst), n):
                yield lst[i : i + n]

        all_completed_prompts = []

        with ThreadPoolExecutor() as executor:
            for batch in tqdm(chunks(results, batch_size), desc="Processing batches"):
                completed_prompts = list(
                    executor.map(
                        lambda result: self.create_prompt(result, rank_start, rank_end),
                        batch,
                    )
                )
                all_completed_prompts.extend(completed_prompts)
        return all_completed_prompts

    def get_num_tokens(self, prompt: str) -> int:
        return len(self._tokenizer.encode(prompt))

    def cost_per_1k_token(self, input_token: bool) -> float:
        return 0

    def get_name(self) -> str:
        return self._name



import logging

from .api_keys import get_azure_openai_args, get_genai_api_key, get_openai_api_key
from .identity_reranker import IdentityReranker
from .rankllm import PromptMode, RankLLM
from .reranker import Reranker

logging.basicConfig(
    level=logging.INFO, format="%(asctime)s - %(levelname)s - %(message)s"
)

__all__ = [
    "IdentityReranker",
    "RankLLM",
    "get_azure_openai_args",
    "get_openai_api_key",
    "get_genai_api_key",
    "PromptMode",
    "Reranker",
]



import unittest
from unittest.mock import MagicMock, patch

from rank_llm.retrieve_and_rerank import retrieve_and_rerank


# Anserini API must be hosted at 8081
class TestRetrieveAndRerank(unittest.TestCase):
    def setUp(self):
        self.patcher = patch("rank_llm.rerank.listwise.rank_listwise_os_llm.load_model")
        self.mock_load_model = self.patcher.start()
        self.mock_llm = MagicMock()
        self.mock_tokenizer = MagicMock()
        self.mock_load_model.return_value = self.mock_llm, self.mock_tokenizer

        self.patcher_cuda = patch("torch.cuda.is_available")
        self.mock_cuda = self.patcher_cuda.start()
        self.mock_cuda.return_value = True

    def tearDown(self):
        self.patcher.stop()
        self.patcher_cuda.stop()

    def test_multiple_datasets(self):
        result = retrieve_and_rerank(
            "unspecified",
            query="chocolate",
            dataset=["msmarco-v2.1-doc", "msmarco-v2.1-doc"],
            interactive=True,
        )

        self.assertEqual(len(result[0]), 2)
        for x in result[0]:
            self.assertEqual(len(x.candidates), 10)

    def test_invalid_retrieval_mode(self):
        # Test handling of invalid retrieval modes
        with self.assertRaises(TypeError):
            retrieve_and_rerank(
                "rank_zephyr",
                ["msmarco-v2.1-doc"],
                interactive=True,
                retrieval_mode="INVALID_MODE",
            )

    def test_improper_arguments(self):
        # Missing dataset
        with self.assertRaises(TypeError):
            retrieve_and_rerank(
                "unspecified", query="hi", interactive=True, num_passes=4
            )
        # Missing model path
        with self.assertRaises(TypeError):
            retrieve_and_rerank(
                query="hi", dataset=["msmarco-v2.1-doc"], interactive=True, num_passes=4
            )
        # Missing query
        with self.assertRaises(TypeError):
            retrieve_and_rerank(
                "rank_zephyr",
                dataset=["msmarco-v2.1-doc"],
                interactive=True,
                num_passes=4,
            )

    def test_multiple_passes(self):
        patcher_rerank_batch = patch(
            "rank_llm.rerank.listwise.RankListwiseOSLLM.rerank_batch", return_value=[]
        )
        mock_rerank_batch = patcher_rerank_batch.start()
        # Test reranking with multiple passes
        retrieve_and_rerank(
            "rank_zephyr",
            query="hi",
            dataset=["msmarco-v2.1-doc"],
            interactive=True,
            num_passes=4,
        )
        # Ensure rerank_batch is called 4 times
        self.assertEqual(mock_rerank_batch.call_count, 4)

        patcher_rerank_batch.stop()

    def test_identity_reranker(self):
        result = retrieve_and_rerank(
            "rank_random",
            query="what is canada",
            dataset=["msmarco-v2.1-doc"],
            top_k_rerank=6,
            interactive=True,
        )
        self.assertEqual(len(result[0][0].candidates), 6)

    def test_identity_reranker_called(self):
        patcher_rerank_batch = patch(
            "rank_llm.rerank.IdentityReranker.rerank_batch", return_value=[]
        )

        for model in ["rank_random", "unspecified"]:
            mock_rerank_batch = patcher_rerank_batch.start()

            retrieve_and_rerank(
                model_path=model,
                query="sample query",
                dataset="msmarco-v2.1-doc",
                top_k_rerank=6,
                interactive=True,
            )
            self.assertTrue(
                mock_rerank_batch.called,
                "IdentityReranker's rerank_batch was not called",
            )

            patcher_rerank_batch.stop()


if __name__ == "__main__":
    unittest.main()



import copy
import logging
import re
from abc import ABC
from datetime import datetime
from functools import cmp_to_key
from typing import Any, Dict, List, Tuple

from ftfy import fix_text
from tqdm import tqdm

from rank_llm.data import Candidate, Request, Result
from rank_llm.rerank.rankllm import PromptMode, RankLLM

logger = logging.getLogger(__name__)


class PairwiseRankLLM(RankLLM, ABC):
    """
    Abstract base class that all pairwise rerankers implement.
    """

    def __init__(
        self,
        model: str,
        context_size: int,
        prompt_mode: PromptMode,
        device: str = "cuda",
        filename: str = "",
        batch_size: int = 32,
    ) -> None:
        super().__init__(model, context_size, prompt_mode)
        self._device = device
        self._filename = filename
        self._batch_size = batch_size

    def rerank_batch(
        self,
        requests: List[Request],
        rank_start: int = 0,
        rank_end: int = 100,
        shuffle_candidates: bool = False,
        logging: bool = False,
        **kwargs: Any,
    ) -> List[Result]:
        """
        Re-rank candidates in a pairwise fashion:
         1. Build a list of all pairwise comparisons.
         2. Process in batches: create prompts, run LLM, update scores.
         3. Sort candidates by final score in descending order.
        """

        rerank_results = [
            Result(
                query=copy.deepcopy(request.query),
                candidates=copy.deepcopy(request.candidates),
                invocations_history=[],
            )
            for request in requests
        ]

        # zero-initialize candidate scores
        for result in rerank_results:
            for candidate in result.candidates:
                candidate.score = 0

        num_queries, num_pairs = len(rerank_results), 0
        self._enumerated_indices = [[] for _ in range(num_queries)]
        for query_idx, res in enumerate(rerank_results):
            num_candidates = len(res.candidates)
            for i in range(num_candidates):
                for j in range(i + 1, num_candidates):
                    self._enumerated_indices[query_idx].append([i, j])
            num_pairs += len(self._enumerated_indices[query_idx])

        with tqdm(
            total=num_pairs, desc="Progress through (q, d) pairs"
        ) as progress_bar:
            for query_idx, pair_list in enumerate(self._enumerated_indices):
                index = 0
                while index < len(pair_list):
                    prompts, token_counts = self.create_prompt_batched(
                        rerank_results, query_idx, index
                    )

                    outputs, output_tokens, scores = self.run_llm_batched(prompts)

                    for (i, j), score in zip(
                        pair_list[index : index + len(scores)], scores
                    ):
                        rerank_results[query_idx].candidates[i].score += score
                        rerank_results[query_idx].candidates[j].score += 1 - score

                    index += self._batch_size
                    progress_bar.update(len(scores))

        for result in rerank_results:
            result.candidates.sort(key=cmp_to_key(self.candidate_comparator))

        return rerank_results

    def create_prompt_batched(
        self, results: List[Result], query_idx: int, index: int
    ) -> Tuple[List[str], List[int]]:
        """
        Create a batch of prompts for the given query_idx, taking pairs of candidates
        from self._enumerated_indices[query_idx] in the range [index : index + batch_size].
        """
        prompts, token_counts = [], []

        pair_list = self._enumerated_indices[query_idx]
        end_index = min(index + self._batch_size, len(pair_list))

        # Build prompts for each pair in [index, end_index)
        for pair_idx in range(index, end_index):
            i, j = pair_list[pair_idx]
            prompt, tcount = self.create_prompt(
                result=results[query_idx], index1=i, index2=j
            )
            prompts.append(prompt)
            token_counts.append(tcount)

        return prompts, token_counts

    def get_output_filename(
        self,
        top_k_candidates: int,
        dataset_name: str,
        shuffle_candidates: bool,
        **kwargs: Any,
    ) -> str:
        if self._filename != "":
            return self._filename
        _modelname = self._model.split("/")[-1]
        if _modelname.startswith("checkpoint"):
            _modelname = self._model.split("/")[-2] + "_" + _modelname
        name = (
            f"{_modelname}_{self._context_size}_{top_k_candidates}_{self._prompt_mode}"
        )
        if dataset_name:
            name = f"{name}_{dataset_name}"

        if shuffle_candidates:
            self._filename = f"{name}_shuffled_{datetime.isoformat(datetime.now())}"
        else:
            self._filename = f"{name}_{datetime.isoformat(datetime.now())}"

        return (
            f"{name}_shuffled_{datetime.isoformat(datetime.now())}"
            if shuffle_candidates
            else f"{name}_{datetime.isoformat(datetime.now())}"
        )

    def candidate_comparator(self, x: Candidate, y: Candidate) -> int:
        if x.score < y.score:
            return 1
        elif x.score > y.score:
            return -1
        else:
            return 0

    def _replace_number(self, s: str) -> str:
        return re.sub(r"\[(\d+)\]", r"(\1)", s)

    def convert_doc_to_prompt_content(
        self, doc: Dict[str, Any], max_length: int
    ) -> str:
        if "text" in doc:
            content = doc["text"]
        elif "segment" in doc:
            content = doc["segment"]
        elif "contents" in doc:
            content = doc["contents"]
        elif "content" in doc:
            content = doc["content"]
        elif "body" in doc:
            content = doc["body"]
        else:
            content = doc["passage"]
        if "title" in doc and doc["title"]:
            content = "Title: " + doc["title"] + " " + "Content: " + content
        content = content.strip()
        content = fix_text(content)
        # For Japanese should cut by character: content = content[:int(max_length)]
        content = " ".join(content.split()[: int(max_length)])
        return self._replace_number(content)



from pathlib import Path
from typing import Any, List, Optional, Tuple

from rank_llm.data import DataWriter, Request, Result
from rank_llm.rerank import (
    PromptMode,
    RankLLM,
    get_azure_openai_args,
    get_genai_api_key,
    get_openai_api_key,
)
from rank_llm.rerank.listwise import RankListwiseOSLLM, SafeGenai, SafeOpenai
from rank_llm.rerank.listwise.rank_fid import RankFiDDistill, RankFiDScore
from rank_llm.rerank.pairwise.duot5 import DuoT5
from rank_llm.rerank.pointwise.monot5 import MonoT5
from rank_llm.rerank.rankllm import RankLLM


class Reranker:
    def __init__(self, model_coordinator: Optional[RankLLM]) -> None:
        self._model_coordinator = model_coordinator

    def rerank_batch(
        self,
        requests: List[Request],
        rank_start: int = 0,
        rank_end: int = 100,
        shuffle_candidates: bool = False,
        logging: bool = False,
        **kwargs: Any,
    ) -> List[Result]:
        """
        Reranks a list of requests using the RankLLM model_coordinator.

        This function applies a sliding window algorithm to rerank the results.
        Each window of results is processed by the RankLLM model_coordinator to obtain a new ranking.

        Args:
            requests (List[Request]): The list of requests. Each request has a query and a candidates list.
            rank_start (int, optional): The starting rank for processing. Defaults to 0.
            rank_end (int, optional): The end rank for processing. Defaults to 100.
            shuffle_candidates (bool, optional): Whether to shuffle candidates before reranking. Defaults to False.
            logging (bool, optional): Enables logging of the reranking process. Defaults to False.
            **kwargs: Additional keyword arguments including:
                populate_invocations_history (bool): Whether to populate the history of inference invocations. Defaults to False.
                window_size (int): The size of the sliding window for listwise reranking, defualts to 20.
                step (int): The size of the step/stride of the sliding window for listwise rernaking, defaults to 10.
                top_k_retrieve (int): The number of retrieved candidates, when set it is used to cap rank_end and window_size.
        Returns:
            List[Result]: A list containing the reranked candidates.
        """
        return self._model_coordinator.rerank_batch(
            requests, rank_start, rank_end, shuffle_candidates, logging, **kwargs
        )

    def rerank(
        self,
        request: Request,
        rank_start: int = 0,
        rank_end: int = 100,
        shuffle_candidates: bool = False,
        logging: bool = False,
        **kwargs: Any,
    ) -> Result:
        """
        Reranks a request using the RankLLM model_coordinator.

        This function applies a sliding window algorithm to rerank the results.
        Each window of results is processed by the RankLLM model_coordinator to obtain a new ranking.

        Args:
            request (Request): The reranking request which has a query and a candidates list.
            rank_start (int, optional): The starting rank for processing. Defaults to 0.
            rank_end (int, optional): The end rank for processing. Defaults to 100.
            shuffle_candidates (bool, optional): Whether to shuffle candidates before reranking. Defaults to False.
            logging (bool, optional): Enables logging of the reranking process. Defaults to False.
            **kwargs: Additional keyword arguments including:
                populate_invocations_history (bool): Whether to populate the history of inference invocations. Defaults to False.
                window_size (int): The size of the sliding window for listwise reranking, defualts to 20.
                step (int): The size of the step/stride of the sliding window for listwise rernaking, defaults to 10.
                top_k_retrieve (int): The number of retrieved candidates, when set it is used to cap rank_end and window size.
        Returns:
            Result: the rerank result which contains the reranked candidates.
        """
        results = self.rerank_batch(
            requests=[request],
            rank_start=rank_start,
            rank_end=rank_end,
            shuffle_candidates=shuffle_candidates,
            logging=logging,
            **kwargs,
        )
        return results[0]

    def write_rerank_results(
        self,
        retrieval_method_name: str,
        results: List[Result],
        shuffle_candidates: bool = False,
        top_k_candidates: int = 100,
        dataset_name: str = None,
        rerank_results_dirname: str = "rerank_results",
        inference_invocations_history_dirname: str = "inference_invocations_history",
        vllm_batched: bool = False,
        sglang_batched: bool = False,
        tensorrt_batched: bool = False,
        **kwargs,
    ) -> str:
        """
        Writes the reranked results to files in specified formats.

        This function saves the reranked results in both TREC Eval format and JSON format.
        The history of inference invocations is saved as well.

        Args:
            retrieval_method_name (str): The name of the retrieval method.
            results (List[Result]): The reranked results to be written.
            shuffle_candidates (bool, optional): Indicates if the candidates were shuffled. Defaults to False.
            top_k_candidates (int, optional): The number of top candidates considered. Defaults to 100.
            pass_ct (int, optional): Pass count, if applicable. Defaults to None.
            window_size (int, optional): The window size used in reranking. Defaults to None.
            dataset_name (str, optional): The name of the dataset used. Defaults to None.
            vllm_batched (bool, optional): Indicates if vLLM inference backend used. Defaults to False.
            sglang_batched (bool, optional): Indicates if SGLang inference backend used. Defaults to False.

        Returns:
            str: The file name of the saved reranked results in TREC Eval format.

        Note:
            The function creates directories and files as needed. The file names are constructed based on the
            provided parameters and the current timestamp to ensure uniqueness so there are no collisions.
        """
        pass_ct: Optional[int] = kwargs.get("pass_ct", None)
        window_size: Optional[int] = kwargs.get("window_size", None)

        name = self._model_coordinator.get_output_filename(
            top_k_candidates, dataset_name, shuffle_candidates, **kwargs
        )

        if window_size is not None:
            name += f"_window_{window_size}"
        if pass_ct is not None:
            name += f"_pass_{pass_ct}"

        # Add vllm or sglang to rerank result file name if they are used
        if vllm_batched:
            name += "_vllm"
        if sglang_batched:
            name += "_sglang"
        if tensorrt_batched:
            name += "_tensorrt"

        # write rerank results
        writer = DataWriter(results)
        Path(f"{rerank_results_dirname}/{retrieval_method_name}/").mkdir(
            parents=True, exist_ok=True
        )
        result_file_name = (
            f"{rerank_results_dirname}/{retrieval_method_name}/{name}.txt"
        )
        writer.write_in_trec_eval_format(result_file_name)
        writer.write_in_jsonl_format(
            f"{rerank_results_dirname}/{retrieval_method_name}/{name}.jsonl"
        )
        # Write the history of inference invocations
        Path(f"{inference_invocations_history_dirname}/{retrieval_method_name}/").mkdir(
            parents=True, exist_ok=True
        )
        writer.write_inference_invocations_history(
            f"{inference_invocations_history_dirname}/{retrieval_method_name}/{name}.json"
        )
        return result_file_name

    def get_model_coordinator(self) -> RankLLM:
        return self._model_coordinator

    def create_model_coordinator(
        model_path: str,
        default_model_coordinator: RankLLM,
        interactive: bool,
        **kwargs: Any,
    ) -> RankLLM:
        """Construct rerank model_coordinator

        Keyword arguments:
        argument -- description
        model_path -- name of model
        default_model_coordinator -- used for interactive mode to pass in a pre-instantiated model_coordinator to use
        interactive -- whether to run retrieve_and_rerank in interactive mode, used by the API

        Return: rerank model_coordinator -- Option<RankLLM>
        """
        use_azure_openai: bool = kwargs.get("use_azure_openai", False)
        vllm_batched: bool = kwargs.get("vllm_batched", False)

        if interactive and default_model_coordinator is not None:
            # Default rerank model_coordinator
            model_coordinator = default_model_coordinator
        elif "gpt" in model_path or use_azure_openai:
            # GPT based reranking models

            keys_and_defaults = [
                ("context_size", 4096),
                ("prompt_mode", PromptMode.RANK_GPT),
                ("num_few_shot_examples", 0),
                ("window_size", 20),
            ]
            [
                context_size,
                prompt_mode,
                num_few_shot_examples,
                window_size,
            ] = extract_kwargs(keys_and_defaults, **kwargs)

            openai_keys = get_openai_api_key()
            model_coordinator = SafeOpenai(
                model=model_path,
                context_size=context_size,
                prompt_mode=prompt_mode,
                window_size=window_size,
                num_few_shot_examples=num_few_shot_examples,
                keys=openai_keys,
                **(get_azure_openai_args() if use_azure_openai else {}),
            )
        elif "gemini" in model_path:
            keys_and_defaults = [
                ("context_size", 4096),
                ("prompt_mode", PromptMode.RANK_GPT),
                ("num_few_shot_examples", 0),
                ("window_size", 20),
            ]
            [
                context_size,
                prompt_mode,
                num_few_shot_examples,
                window_size,
            ] = extract_kwargs(keys_and_defaults, **kwargs)

            genai_keys = get_genai_api_key()
            model_coordinator = SafeGenai(
                model=model_path,
                context_size=context_size,
                prompt_mode=prompt_mode,
                num_few_shot_examples=num_few_shot_examples,
                window_size=window_size,
                keys=genai_keys,
            )

        elif "vicuna" in model_path or "zephyr" in model_path:
            # RankVicuna or RankZephyr model suite
            print(f"Loading {model_path} ...")

            model_full_paths = {
                "rank_zephyr": "castorini/rank_zephyr_7b_v1_full",
                "rank_vicuna": "castorini/rank_vicuna_7b_v1",
            }

            keys_and_defaults = [
                ("context_size", 4096),
                ("prompt_mode", PromptMode.RANK_GPT),
                ("num_few_shot_examples", 0),
                ("device", "cuda"),
                ("num_gpus", 1),
                ("variable_passages", False),
                ("window_size", 20),
                ("system_message", None),
                ("vllm_batched", False),
                ("sglang_batched", False),
                ("tensorrt_batched", False),
                ("use_logits", False),
                ("use_alpha", False),
            ]
            [
                context_size,
                prompt_mode,
                num_few_shot_examples,
                device,
                num_gpus,
                variable_passages,
                window_size,
                system_message,
                vllm_batched,
                sglang_batched,
                tensorrt_batched,
                use_logits,
                use_alpha,
            ] = extract_kwargs(keys_and_defaults, **kwargs)

            model_coordinator = RankListwiseOSLLM(
                model=(
                    model_full_paths[model_path]
                    if model_path in model_full_paths
                    else model_path
                ),
                name=model_path,
                context_size=context_size,
                prompt_mode=prompt_mode,
                num_few_shot_examples=num_few_shot_examples,
                device=device,
                num_gpus=num_gpus,
                variable_passages=variable_passages,
                window_size=window_size,
                system_message=system_message,
                vllm_batched=vllm_batched,
                sglang_batched=sglang_batched,
                tensorrt_batched=tensorrt_batched,
                use_logits=use_logits,
                use_alpha=use_alpha,
            )

            print(f"Completed loading {model_path}")
        elif "monot5" in model_path:
            # using monot5
            print(f"Loading {model_path} ...")

            model_full_paths = {"monot5": "castorini/monot5-3b-msmarco-10k"}

            keys_and_defaults = [
                ("prompt_mode", PromptMode.MONOT5),
                ("context_size", 512),
                ("device", "cuda"),
                ("batch_size", 64),
            ]
            [prompt_mode, context_size, device, batch_size] = extract_kwargs(
                keys_and_defaults, **kwargs
            )

            model_coordinator = MonoT5(
                model=(
                    model_full_paths[model_path]
                    if model_path in model_full_paths
                    else model_path
                ),
                prompt_mode=prompt_mode,
                context_size=context_size,
                device=device,
                batch_size=batch_size,
            )
        elif "duot5" in model_path:
            # using monot5
            print(f"Loading {model_path} ...")

            model_full_paths = {"duot5": "castorini/duot5-3b-msmarco-10k"}

            keys_and_defaults = [
                ("prompt_mode", PromptMode.DUOT5),
                ("context_size", 512),
                ("device", "cuda"),
                ("batch_size", 64),
            ]
            [prompt_mode, context_size, device, batch_size] = extract_kwargs(
                keys_and_defaults, **kwargs
            )

            model_coordinator = DuoT5(
                model=(
                    model_full_paths[model_path]
                    if model_path in model_full_paths
                    else model_path
                ),
                prompt_mode=prompt_mode,
                context_size=context_size,
                device=device,
                batch_size=batch_size,
            )
        elif "lit5-distill" in model_path.lower():
            keys_and_defaults = [
                ("context_size", 150),
                ("prompt_mode", PromptMode.LiT5),
                ("num_few_shot_examples", 0),
                ("window_size", 20),
                ("precision", "bfloat16"),
                ("device", "cuda"),
                # reuse this parameter, but its not for "vllm", but only for "batched"
                ("vllm_batched", False),
            ]
            (
                context_size,
                prompt_mode,
                num_few_shot_examples,
                window_size,
                precision,
                device,
                vllm_batched,
            ) = extract_kwargs(keys_and_defaults, **kwargs)

            model_coordinator = RankFiDDistill(
                model=model_path,
                context_size=context_size,
                prompt_mode=prompt_mode,
                num_few_shot_examples=num_few_shot_examples,
                window_size=window_size,
                precision=precision,
                device=device,
                batched=vllm_batched,
            )
            print(f"Completed loading {model_path}")
        elif "lit5-score" in model_path.lower():
            keys_and_defaults = [
                ("context_size", 150),
                ("prompt_mode", PromptMode.LiT5),
                ("num_few_shot_examples", 0),
                ("window_size", 100),
                ("precision", "bfloat16"),
                ("device", "cuda"),
                # reuse this parameter, but its not for "vllm", but only for "batched"
                ("vllm_batched", False),
            ]
            (
                context_size,
                prompt_mode,
                num_few_shot_examples,
                window_size,
                precision,
                device,
                vllm_batched,
            ) = extract_kwargs(keys_and_defaults, **kwargs)

            model_coordinator = RankFiDScore(
                model=model_path,
                context_size=context_size,
                prompt_mode=prompt_mode,
                num_few_shot_examples=num_few_shot_examples,
                window_size=window_size,
                precision=precision,
                device=device,
                batched=vllm_batched,
            )
            print(f"Completed loading {model_path}")
        elif vllm_batched:
            # supports loading models from huggingface
            print(f"Loading {model_path} ...")
            keys_and_defaults = [
                ("context_size", 4096),
                ("prompt_mode", PromptMode.RANK_GPT),
                ("num_few_shot_examples", 0),
                ("device", "cuda"),
                ("num_gpus", 1),
                ("variable_passages", False),
                ("window_size", 20),
                ("system_message", None),
                ("vllm_batched", True),
                ("use_logits", False),
                ("use_alpha", False),
            ]
            [
                context_size,
                prompt_mode,
                num_few_shot_examples,
                device,
                num_gpus,
                variable_passages,
                window_size,
                system_message,
                vllm_batched,
                use_logits,
                use_alpha,
            ] = extract_kwargs(keys_and_defaults, **kwargs)

            model_coordinator = RankListwiseOSLLM(
                model=(model_path),
                name=model_path,
                context_size=context_size,
                prompt_mode=prompt_mode,
                num_few_shot_examples=num_few_shot_examples,
                device=device,
                num_gpus=num_gpus,
                variable_passages=variable_passages,
                window_size=window_size,
                system_message=system_message,
                use_logits=use_logits,
                use_alpha=use_alpha,
                vllm_batched=vllm_batched,
            )

            print(f"Completed loading {model_path}")
        elif model_path in ["unspecified", "rank_random", "rank_identity"]:
            # NULL reranker
            model_coordinator = None
        else:
            raise ValueError(f"Unsupported model: {model_path}")

        if model_coordinator is None and model_path not in [
            "unspecified",
            "rank_random",
            "rank_identity",
        ]:
            raise ValueError(f"Unsupported model: {model_path}")
        return model_coordinator


def extract_kwargs(
    keys_and_defaults: List[Tuple[str, Any]],
    **kwargs,
) -> List[Any]:
    """Extract specified kwargs from **kwargs

    Keyword arguments:
    keys_and_defaults -- List of Tuple(keyname, default)
    Return: List of extracted kwargs in order provided in keys_and_defaults
    """
    extracted_kwargs = []
    for key, default in keys_and_defaults:
        value = kwargs.get(key, None)
        if value is None:
            value = default
        if (
            value is not None
            and default is not None
            and not isinstance(value, type(default))
        ):
            raise ValueError(
                f"Provided kwarg for {key} must be of type {type(default).__name__}, got {type(value).__name__}"
            )
        extracted_kwargs.append(value)
    return extracted_kwargs



from urllib import parse

import requests

from rank_llm.data import Candidate, Query, Request

from . import RetrievalMethod, RetrievalMode


class ServiceRetriever:
    def __init__(
        self,
        retrieval_mode: RetrievalMode = RetrievalMode.DATASET,
        retrieval_method: RetrievalMethod = RetrievalMethod.BM25,
    ) -> None:
        """
        Creates a ServiceRetriever instance with a specified retrieval method and mode.

        Args:
            retrieval_mode (RetrievalMode): The retrieval mode to be used. Defaults to DATASET. Only DATASET mode is currently supported.
            retrieval_method (RetrievalMethod): The retrieval method to be used. Defaults to BM25.

        Raises:
            ValueError: If retrieval mode or retrieval method is invalid or missing.
        """
        self._retrieval_mode = retrieval_mode
        self._retrieval_method = retrieval_method

        if retrieval_mode != RetrievalMode.DATASET:
            raise ValueError(
                f"{retrieval_mode} is not supported for ServiceRetriever. Only DATASET mode is currently supported."
            )

        if retrieval_method != RetrievalMethod.BM25:
            raise ValueError(
                f"{retrieval_method} is not supported for ServiceRetriever. Only BM25 is currently supported."
            )

    def retrieve(
        self,
        dataset: str,
        request: Request,
        k: int = 50,
        host: str = "http://localhost:8081",
        timeout: int = 15
        * 60,  # downloding and decompressing the index can take a long time.
    ) -> Request:
        """
        Executes the retrieval process based on the configation provided with the Retriever instance. Takes in a Request object with a query and empty candidates object and the top k items to retrieve.

        Args:
            request (Request): The request containing the query and qid.
            dataset (str): The name of the dataset.
            k (int, optional): The top k hits to retrieve. Defaults to 100.
            host (str): The Anserini API host address. Defaults to http://localhost:8081

        Returns:
            Request. Contains a query and list of candidates
        Raises:
            ValueError: If the retrieval mode is invalid or the result format is not as expected.
        """

        url = f"{host}/api/v1.0/indexes/{dataset}/search?query={parse.quote(request.query.text)}&hits={str(k)}&qid={request.query.qid}"
        print(url)
        try:
            response = requests.get(url, timeout=timeout)
            response.raise_for_status()
        except requests.exceptions.RequestException as e:
            raise type(e)(
                f"Failed to retrieve data from Anserini server: {str(e)}"
            ) from e

        data = response.json()
        retrieved_results = Request(
            query=Query(text=data["query"]["text"], qid=data["query"]["qid"])
        )

        for candidate in data["candidates"]:
            retrieved_results.candidates.append(
                Candidate(
                    docid=candidate["docid"],
                    score=candidate["score"],
                    doc=candidate["doc"],
                )
            )

        return retrieved_results



import hashlib
import logging
import os
import re
from urllib.error import HTTPError, URLError
from urllib.request import urlretrieve

from tqdm import tqdm

from rank_llm.retrieve.repo_info import HITS_INFO

logger = logging.getLogger(__name__)


# https://gist.github.com/leimao/37ff6e990b3226c2c9670a2cd1e4a6f5
class TqdmUpTo(tqdm):
    def update_to(self, b=1, bsize=1, tsize=None):
        """
        b  : int, optional
            Number of blocks transferred so far [default: 1].
        bsize  : int, optional
            Size of each block (in tqdm units) [default: 1].
        tsize  : int, optional
            Total size (in tqdm units). If [default: None] remains unchanged.
        """
        if tsize is not None:
            self.total = tsize
        self.update(b * bsize - self.n)  # will also set self.n = b * bsize


# For large files, we need to compute MD5 block by block. See:
# https://stackoverflow.com/questions/1131220/get-md5-hash-of-big-files-in-python
def compute_md5(file, block_size=2**20):
    m = hashlib.md5()
    with open(file, "rb") as f:
        while True:
            buf = f.read(block_size)
            if not buf:
                break
            m.update(buf)
    return m.hexdigest()


def download_url(
    url, save_dir, local_filename=None, md5=None, force=False, verbose=True
):
    # If caller does not specify local filename, figure it out from the download URL:
    if not local_filename:
        filename = url.split("/")[-1]
        filename = re.sub(
            "\\?dl=1$", "", filename
        )  # Remove the Dropbox 'force download' parameter
    else:
        # Otherwise, use the specified local_filename:
        filename = local_filename

    destination_path = os.path.join(save_dir, filename)

    if verbose:
        print(f"curr_path{os.getcwd()}")
        print(f"Downloading {url} to {destination_path}...")

    # Check to see if file already exists, if so, simply return (quietly) unless force=True, in which case we remove
    # destination file and download fresh copy.
    if os.path.exists(destination_path):
        if verbose:
            print(f"{destination_path} already exists!")
        if not force:
            if verbose:
                print(f"Skipping download.")
            return destination_path
        if verbose:
            print(f"force=True, removing {destination_path}; fetching fresh copy...")
        os.remove(destination_path)

    with TqdmUpTo(
        unit="B", unit_scale=True, unit_divisor=1024, miniters=1, desc=filename
    ) as t:
        urlretrieve(url, filename=destination_path, reporthook=t.update_to)

    if md5:
        md5_computed = compute_md5(destination_path)
        assert (
            md5_computed == md5
        ), f"{destination_path} does not match checksum! Expecting {md5} got {md5_computed}."

    return destination_path


def get_cache_home():
    custom_dir = os.environ.get("RANK_LLM_CACHE")
    if custom_dir is not None and custom_dir != "":
        print("custom")
        return custom_dir
    return os.path.expanduser(
        os.path.join(f"~{os.path.sep}rank_llm", "retrieve_results")
    )


def download_and_unpack_hits(
    url,
    hits_directory="files",
    local_filename=False,
    force=False,
    verbose=True,
    prebuilt=False,
    md5=None,
):
    # If caller does not specify local filename, figure it out from the download URL:
    if not local_filename:
        file_name = url.split("/")[-1]
    else:
        # Otherwise, use the specified local_filename:
        file_name = local_filename

    if prebuilt:
        hits_directory = os.path.join(get_cache_home(), hits_directory)
        file_path = os.path.join(hits_directory, f"{file_name}.{md5}")

        if not os.path.exists(hits_directory):
            os.makedirs(hits_directory)

        local_file = os.path.join(hits_directory, file_name)
    else:
        local_file = os.path.join(hits_directory, file_name)
        file_path = os.path.join(hits_directory, file_name)

    # Check to see if file already exists, if so, simply return (quietly) unless force=True, in which case we remove
    # file and download fresh copy.
    if os.path.exists(file_path):
        if not force:
            if verbose:
                print(f"{file_path} already exists, skipping download.")
            return file_path
        if verbose:
            print(
                f"{file_path} already exists, but force=True, removing {file_path} and fetching fresh copy..."
            )
        os.remove(file_path)

    print(f"Downloading file at {url}...")
    download_url(
        url, hits_directory, local_filename=local_filename, verbose=False, md5=md5
    )

    # No need to extract for JSON and text files
    if verbose:
        print(f"File {local_file} has been downloaded to {file_path}.")
    return local_file


def download_cached_hits(query_name, force=False, verbose=True, mirror=None):
    if query_name not in HITS_INFO:
        print(f"query_name unrecognized {query_name}")
        raise ValueError(f"Unrecognized query name {query_name}")
    query_md5 = HITS_INFO[query_name]["md5"]
    for url in HITS_INFO[query_name]["urls"]:
        print(f"Trying to download cached retrieved hits at {url}...")
        try:
            hits_dir = query_name.rsplit("/", 2)[-2]
            return download_and_unpack_hits(
                url, hits_directory=hits_dir, prebuilt=True, md5=query_md5
            )
        except (HTTPError, URLError) as e:
            print(
                f"Unable to download cached retrieved hits at {url}, trying next URL..."
            )
    raise ValueError(f"Unable to download cached retrieved hits at any known URLs.")


get_cache_home()



from .indices_dict import INDICES
from .pyserini_retriever import PyseriniRetriever, RetrievalMethod
from .repo_info import HITS_INFO
from .retriever import RetrievalMode, Retriever
from .service_retriever import ServiceRetriever
from .topics_dict import TOPICS
from .utils import (
    compute_md5,
    download_and_unpack_hits,
    download_cached_hits,
    download_url,
)

__all__ = [
    "INDICES",
    "TOPICS",
    "HITS_INFO",
    "RetrievalMethod",
    "PyseriniRetriever",
    "ServiceRetriever",
    "Retriever",
    "RetrievalMode",
    "download_and_unpack_hits",
    "download_cached_hits",
    "download_url",
    "compute_md5",
]



from .rank_gemini import SafeGenai
from .rank_gpt import SafeOpenai
from .rank_listwise_os_llm import RankListwiseOSLLM
from .vicuna_reranker import VicunaReranker
from .zephyr_reranker import ZephyrReranker

__all__ = [
    "RankListwiseOSLLM",
    "VicunaReranker",
    "ZephyrReranker",
    "SafeOpenai",
    "SafeGenai",
]



HITS_INFO = {
    "BM25/retrieve_results_covid_top100.jsonl": {
        "description": "Top-100 BM25 Results for the TREC COVID task. (Cached JSONL format)",
        "urls": [
            "https://github.com/castorini/rank_llm_data/raw/main/retrieve_results/BM25/retrieve_results_covid_top100.jsonl",
        ],
        "md5": "7dd70b911afc40647c37580e2f17a27f",
        "size (bytes)": 8779956,
        "downloaded": False,
    },
    "BM25/retrieve_results_dl19_top100.jsonl": {
        "description": "Top-100 BM25 Results for the TREC DL19 task. (Cached JSONL format)",
        "urls": [
            "https://github.com/castorini/rank_llm_data/raw/main/retrieve_results/BM25/retrieve_results_dl19_top100.jsonl",
        ],
        "md5": "518ab9f821fb74d862589d1d141c208e",
        "size (bytes)": 2000369,
        "downloaded": False,
    },
    "BM25/retrieve_results_dl20_top100.jsonl": {
        "description": "Top-100 BM25 Results for the TREC DL20 task. (Cached JSONL format)",
        "urls": [
            "https://github.com/castorini/rank_llm_data/raw/main/retrieve_results/BM25/retrieve_results_dl20_top100.jsonl",
        ],
        "md5": "3b6397f50fd78229e8d733243745a6d4",
        "size (bytes)": 2528494,
        "downloaded": False,
    },
    "BM25/retrieve_results_dl21_top100.jsonl": {
        "description": "Top-100 BM25 Results for the TREC DL21 task. (Cached JSONL format)",
        "urls": [
            "https://github.com/castorini/rank_llm_data/raw/main/retrieve_results/BM25/retrieve_results_dl21_top100.jsonl",
        ],
        "md5": "6ca5e405c55fc2280d66eb0afc7416f5",
        "size (bytes)": 2140373,
        "downloaded": False,
    },
    "BM25/retrieve_results_dl22_top100.jsonl": {
        "description": "Top-100 BM25 Results for the TREC DL22 task. (Cached JSONL format)",
        "urls": [
            "https://github.com/castorini/rank_llm_data/raw/main/retrieve_results/BM25/retrieve_results_dl22_top100.jsonl",
        ],
        "md5": "08322ee30bf084d1309f5c5a515f2f18",
        "size (bytes)": 3021942,
        "downloaded": False,
    },
    "BM25/retrieve_results_msmarco-v2.1-doc-segmented.bm25.rag24.raggy-dev_top100.jsonl": {
        "description": "Top-100 BM25 Results for the TREC24-Raggy Dev task. (Cached JSONL format)",
        "urls": [
            "https://github.com/castorini/rank_llm_data/raw/main/retrieve_results/BM25/retrieve_results_msmarco-v2.1-doc-segmented.bm25.rag24.raggy-dev_top100.jsonl",
        ],
        "md5": "ff8a8db2180bc628115d678135179a21",
        "size (bytes)": 26911601,
        "downloaded": False,
    },
    "BM25/retrieve_results_msmarco-v2.1-doc-segmented.bm25.rag24.researchy-dev_top100.jsonl": {
        "description": "Top-100 BM25 Results for the TREC24-Researchy Dev task. (Cached JSONL format)",
        "urls": [
            "https://github.com/castorini/rank_llm_data/raw/main/retrieve_results/BM25/retrieve_results_msmarco-v2.1-doc-segmented.bm25.rag24.researchy-dev_top100.jsonl",
        ],
        "md5": "41ff53ef4235f94924c75bb35f46f5f7",
        "size (bytes)": 141212740,
        "downloaded": False,
    },
    "BM25/retrieve_results_news_top100.jsonl": {
        "description": "Top-100 BM25 Results for the TREC News task. (Cached JSONL format)",
        "urls": [
            "https://github.com/castorini/rank_llm_data/raw/main/retrieve_results/BM25/retrieve_results_news_top100.jsonl",
        ],
        "md5": "3fa0063db9ba1dcfc5b527a062d41720",
        "size (bytes)": 41110970,
        "downloaded": False,
    },
    "BM25/retrieve_results_noveleval_top100.jsonl": {
        "description": "Top-100 BM25 Results for the NovelEval task. (Cached JSONL format)",
        "urls": [
            "https://github.com/castorini/rank_llm_data/raw/main/retrieve_results/BM25/retrieve_results_noveleval_top100.jsonl",
        ],
        "md5": "96b9e6bc7e881122ab11a94656da816e",
        "size (bytes)": 417977,
        "downloaded": False,
    },
    "BM25/trec_results_covid_top100.txt": {
        "description": "Top-100 BM25 Results for the TREC COVID task. (TREC format)",
        "urls": [
            "https://github.com/castorini/rank_llm_data/raw/main/retrieve_results/retrieve_results/BM25/trec_results_covid_top100.txt",
        ],
        "md5": "ba0360f2abe60abfe361665942fbdc4b",
        "size (bytes)": 204370,
        "downloaded": False,
    },
    "BM25/trec_results_dl19_top100.txt": {
        "description": "Top-100 BM25 Results for the TREC DL19 task. (TREC format)",
        "urls": [
            "https://github.com/castorini/rank_llm_data/raw/main/retrieve_results/retrieve_results/BM25/trec_results_dl19_top100.txt",
        ],
        "md5": "4be11ccaf5808bd861b831b1df960907",
        "size (bytes)": 190529,
        "downloaded": False,
    },
    "BM25/trec_results_dl20_top100.txt": {
        "description": "Top-100 BM25 Results for the TREC DL20 task. (TREC format)",
        "urls": [
            "https://github.com/castorini/rank_llm_data/raw/main/retrieve_results/retrieve_results/BM25/trec_results_dl20_top100.txt",
        ],
        "md5": "63d07d941edc44430b3ddec3a65a0433",
        "size (bytes)": 240316,
        "downloaded": False,
    },
    "BM25/trec_results_dl21_top100.txt": {
        "description": "Top-100 BM25 Results for the TREC DL21 task. (TREC format)",
        "urls": [
            "https://github.com/castorini/rank_llm_data/raw/main/retrieve_results/retrieve_results/BM25/trec_results_dl21_top100.txt",
        ],
        "md5": "900915b48cbbd25b87f8cf0bfe36496f",
        "size (bytes)": 347951,
        "downloaded": False,
    },
    "BM25/trec_results_dl22_top100.txt": {
        "description": "Top-100 BM25 Results for the TREC DL22 task. (TREC format)",
        "urls": [
            "https://github.com/castorini/rank_llm_data/raw/main/retrieve_results/retrieve_results/BM25/trec_results_dl22_top100.txt",
        ],
        "md5": "4e93d4a41ee9343fda24887534cc0943",
        "size (bytes)": 503425,
        "downloaded": False,
    },
    "BM25/trec_results_msmarco-v2.1-doc-segmented.bm25.rag24.raggy-dev_top100.txt": {
        "description": "Top-100 BM25 Results for the TREC24-Raggy Dev task. (TREC format)",
        "urls": [
            "https://github.com/castorini/rank_llm_data/raw/main/retrieve_results/retrieve_results/BM25/trec_results_msmarco-v2.1-doc-segmented.bm25.rag24.raggy-dev_top100.txt",
        ],
        "md5": "2e332446044fc6fe1197c02ddaebaab8",
        "size (bytes)": 910820,
        "downloaded": False,
    },
    "BM25/trec_results_msmarco-v2.1-doc-segmented.bm25.rag24.researchy-dev_top100.txt": {
        "description": "Top-100 BM25 Results for the TREC24-Researchy Dev task. (TREC format)",
        "urls": [
            "https://github.com/castorini/rank_llm_data/raw/main/retrieve_results/retrieve_results/BM25/trec_results_msmarco-v2.1-doc-segmented.bm25.rag24.researchy-dev_top100.txt",
        ],
        "md5": "f90bf6b4a37902cd34ce1113886fd209",
        "size (bytes)": 4496144,
        "downloaded": False,
    },
    "BM25/trec_results_news_top100.txt": {
        "description": "Top-100 BM25 Results for the TREC News task. (TREC format)",
        "urls": [
            "https://github.com/castorini/rank_llm_data/raw/main/retrieve_results/retrieve_results/BM25/trec_results_news_top100.txt",
        ],
        "md5": "33e2fe3e5cd7332f0933b48516fbbe48",
        "size (bytes)": 390316,
        "downloaded": False,
    },
    "BM25_RM3/retrieve_results_covid_top100.jsonl": {
        "description": "Top-100 BM25 + RM3 Results for the TREC COVID task. (Cached JSONL format)",
        "urls": [
            "https://github.com/castorini/rank_llm_data/raw/main/retrieve_results/BM25_RM3/retrieve_results_covid_top100.jsonl",
        ],
        "md5": "f80e8341da7de1d0cce166ed6e2fa84e",
        "size (bytes)": 9020100,
        "downloaded": False,
    },
    "BM25_RM3/retrieve_results_dl19_top100.jsonl": {
        "description": "Top-100 BM25 + RM3 Results for the TREC DL19 task. (Cached JSONL format)",
        "urls": [
            "https://github.com/castorini/rank_llm_data/raw/main/retrieve_results/BM25_RM3/retrieve_results_dl19_top100.jsonl",
        ],
        "md5": "2f8ff225e60baa2c2d47162ae4d371c1",
        "size (bytes)": 2054706,
        "downloaded": False,
    },
    "BM25_RM3/retrieve_results_dl20_top100.jsonl": {
        "description": "Top-100 BM25 + RM3 Results for the TREC DL20 task. (Cached JSONL format)",
        "urls": [
            "https://github.com/castorini/rank_llm_data/raw/main/retrieve_results/BM25_RM3/retrieve_results_dl20_top100.jsonl",
        ],
        "md5": "b8414ab263a05df678d375281c2bc000",
        "size (bytes)": 2548091,
        "downloaded": False,
    },
    "BM25_RM3/retrieve_results_dl21_top100.jsonl": {
        "description": "Top-100 BM25 + RM3 Results for the TREC DL21 task. (Cached JSONL format)",
        "urls": [
            "https://github.com/castorini/rank_llm_data/raw/main/retrieve_results/BM25_RM3/retrieve_results_dl21_top100.jsonl",
        ],
        "md5": "1a0841bb0f82b820bfb2de700a9e89b5",
        "size (bytes)": 2157834,
        "downloaded": False,
    },
    "BM25_RM3/retrieve_results_dl22_top100.jsonl": {
        "description": "Top-100 BM25 + RM3 Results for the TREC DL22 task. (Cached JSONL format)",
        "urls": [
            "https://github.com/castorini/rank_llm_data/raw/main/retrieve_results/BM25_RM3/retrieve_results_dl22_top100.jsonl",
        ],
        "md5": "1dd01bb2bb44b41d6af77c2879492be7",
        "size (bytes)": 3060472,
        "downloaded": False,
    },
    "BM25_RM3/retrieve_results_news_top100.jsonl": {
        "description": "Top-100 BM25 + RM3 Results for the TREC News task. (Cached JSONL format)",
        "urls": [
            "https://github.com/castorini/rank_llm_data/raw/main/retrieve_results/BM25_RM3/retrieve_results_news_top100.jsonl",
        ],
        "md5": "c5e8733df0a5276df2fb5b0862dba0b3",
        "size (bytes)": 38912304,
        "downloaded": False,
    },
    "BM25_RM3/trec_results_covid_top100.txt": {
        "description": "Top-100 BM25 + RM3 Results for the TREC COVID task. (TREC format)",
        "urls": [
            "https://github.com/castorini/rank_llm_data/raw/main/retrieve_results/retrieve_results/BM25_RM3/trec_results_covid_top100.txt",
        ],
        "md5": "a15d25b72b0a17b728f8215b704d7e36",
        "size (bytes)": 207451,
        "downloaded": False,
    },
    "BM25_RM3/trec_results_dl19_top100.txt": {
        "description": "Top-100 BM25 + RM3 Results for the TREC DL19 task. (TREC format)",
        "urls": [
            "https://github.com/castorini/rank_llm_data/raw/main/retrieve_results/retrieve_results/BM25_RM3/trec_results_dl19_top100.txt",
        ],
        "md5": "d2d385f402ccea28f0b4b43f4d90b2af",
        "size (bytes)": 191863,
        "downloaded": False,
    },
    "BM25_RM3/trec_results_dl20_top100.txt": {
        "description": "Top-100 BM25 + RM3 Results for the TREC DL20 task. (TREC format)",
        "urls": [
            "https://github.com/castorini/rank_llm_data/raw/main/retrieve_results/retrieve_results/BM25_RM3/trec_results_dl20_top100.txt",
        ],
        "md5": "de31f7400970d0472bdeaba6717c0876",
        "size (bytes)": 241312,
        "downloaded": False,
    },
    "BM25_RM3/trec_results_dl21_top100.txt": {
        "description": "Top-100 BM25 + RM3 Results for the TREC DL21 task. (TREC format)",
        "urls": [
            "https://github.com/castorini/rank_llm_data/raw/main/retrieve_results/retrieve_results/BM25_RM3/trec_results_dl21_top100.txt",
        ],
        "md5": "18dd1d0be4d5126be665d48449acde02",
        "size (bytes)": 347048,
        "downloaded": False,
    },
    "BM25_RM3/trec_results_dl22_top100.txt": {
        "description": "Top-100 BM25 + RM3 Results for the TREC DL22 task. (TREC format)",
        "urls": [
            "https://github.com/castorini/rank_llm_data/raw/main/retrieve_results/retrieve_results/BM25_RM3/trec_results_dl22_top100.txt",
        ],
        "md5": "071d13dab7b88ec2299232ecbb525dd9",
        "size (bytes)": 503886,
        "downloaded": False,
    },
    "BM25_RM3/trec_results_news_top100.txt": {
        "description": "Top-100 BM25 + RM3 Results for the TREC News task. (TREC format)",
        "urls": [
            "https://github.com/castorini/rank_llm_data/raw/main/retrieve_results/retrieve_results/BM25_RM3/trec_results_news_top100.txt",
        ],
        "md5": "9838a6b5c067edcfeb80bf04222b0c87",
        "size (bytes)": 391639,
        "downloaded": False,
    },
    "D_BERT_KD_TASB/retrieve_results_dl19_top100.jsonl": {
        "description": "Top-100 D_BERT_KD_TASB Results for the TREC DL19 task. (Cached JSONL format)",
        "urls": [
            "https://github.com/castorini/rank_llm_data/raw/main/retrieve_results/D_BERT_KD_TASB/retrieve_results_dl19_top100.jsonl",
        ],
        "md5": "22f14897bf37098aa4836e254287b2f1",
        "size (bytes)": 1864083,
        "downloaded": False,
    },
    "D_BERT_KD_TASB/retrieve_results_dl20_top100.jsonl": {
        "description": "Top-100 D_BERT_KD_TASB Results for the TREC DL20 task. (Cached JSONL format)",
        "urls": [
            "https://github.com/castorini/rank_llm_data/raw/main/retrieve_results/D_BERT_KD_TASB/retrieve_results_dl20_top100.jsonl",
        ],
        "md5": "9221b308063d3cd42c574e7dbff38941",
        "size (bytes)": 2318086,
        "downloaded": False,
    },
    "D_BERT_KD_TASB/trec_results_dl19_top100.txt": {
        "description": "Top-100 D_BERT_KD_TASB Results for the TREC DL19 task. (TREC format)",
        "urls": [
            "https://github.com/castorini/rank_llm_data/raw/main/retrieve_results/retrieve_results/D_BERT_KD_TASB/trec_results_dl19_top100.txt",
        ],
        "md5": "8f107e6075eff57bb56e964aec2ee10a",
        "size (bytes)": 190741,
        "downloaded": False,
    },
    "D_BERT_KD_TASB/trec_results_dl20_top100.txt": {
        "description": "Top-100 D_BERT_KD_TASB Results for the TREC DL20 task. (TREC format)",
        "urls": [
            "https://github.com/castorini/rank_llm_data/raw/main/retrieve_results/retrieve_results/D_BERT_KD_TASB/trec_results_dl20_top100.txt",
        ],
        "md5": "9456b813de73fb68e2e12222ef9df8c4",
        "size (bytes)": 240082,
        "downloaded": False,
    },
    "OPEN_AI_ADA2/retrieve_results_dl19_top100.jsonl": {
        "description": "Top-100 OpenAI ADA2 Results for the TREC DL19 task. (Cached JSONL format)",
        "urls": [
            "https://github.com/castorini/rank_llm_data/raw/main/retrieve_results/OPEN_AI_ADA2/retrieve_results_dl19_top100.jsonl",
        ],
        "md5": "16d5a9a1448b58d2f89c844a7c6b4a4c",
        "size (bytes)": 1849150,
        "downloaded": False,
    },
    "OPEN_AI_ADA2/retrieve_results_dl20_top100.jsonl": {
        "description": "Top-100 OpenAI ADA2 Results for the TREC DL20 task. (Cached JSONL format)",
        "urls": [
            "https://github.com/castorini/rank_llm_data/raw/main/retrieve_results/OPEN_AI_ADA2/retrieve_results_dl20_top100.jsonl",
        ],
        "md5": "7480ae59f7146241e13ff39efb47d3f3",
        "size (bytes)": 2306232,
        "downloaded": False,
    },
    "OPEN_AI_ADA2/trec_results_dl19_top100.txt": {
        "description": "Top-100 OpenAI ADA2 Results for the TREC DL19 task. (TREC format)",
        "urls": [
            "https://github.com/castorini/rank_llm_data/raw/main/retrieve_results/retrieve_results/OPEN_AI_ADA2/trec_results_dl19_top100.txt",
        ],
        "md5": "63052f75197a6e01b86165d4d971f467",
        "size (bytes)": 193174,
        "downloaded": False,
    },
    "OPEN_AI_ADA2/trec_results_dl20_top100.txt": {
        "description": "Top-100 OpenAI ADA2 Results for the TREC DL20 task. (TREC format)",
        "urls": [
            "https://github.com/castorini/rank_llm_data/raw/main/retrieve_results/retrieve_results/OPEN_AI_ADA2/trec_results_dl20_top100.txt",
        ],
        "md5": "89482187c1cd5cc38b5288f4fd6c7369",
        "size (bytes)": 243029,
        "downloaded": False,
    },
    "REP_LLAMA/retrieve_results_dl19_top100.jsonl": {
        "description": "Top-100 repLLaMA Results for the TREC DL19 task. (Cached JSONL format)",
        "urls": [
            "https://github.com/castorini/rank_llm_data/raw/main/retrieve_results/REP_LLAMA/retrieve_results_dl19_top100.jsonl",
        ],
        "md5": "f3cd43822547ade6602dd12af9144ebc",
        "size (bytes)": 3648493,
        "downloaded": False,
    },
    "REP_LLAMA/retrieve_results_dl20_top100.jsonl": {
        "description": "Top-100 repLLaMA Results for the TREC DL20 task. (Cached JSONL format)",
        "urls": [
            "https://github.com/castorini/rank_llm_data/raw/main/retrieve_results/REP_LLAMA/retrieve_results_dl20_top100.jsonl",
        ],
        "md5": "e994e98c1d1f495a8a311670d5df7f07",
        "size (bytes)": 4581491,
        "downloaded": False,
    },
    "REP_LLAMA/trec_results_dl19_top100.txt": {
        "description": "Top-100 repLLaMA Results for the TREC DL19 task. (TREC format)",
        "urls": [
            "https://github.com/castorini/rank_llm_data/raw/main/retrieve_results/retrieve_results/REP_LLAMA/trec_results_dl19_top100.txt",
        ],
        "md5": "72adfd94f7fb04b6ae57674034cfa4e4",
        "size (bytes)": 1415843,
        "downloaded": False,
    },
    "REP_LLAMA/trec_results_dl20_top100.txt": {
        "description": "Top-100 repLLaMA Results for the TREC DL20 task. (TREC format)",
        "urls": [
            "https://github.com/castorini/rank_llm_data/raw/main/retrieve_results/retrieve_results/REP_LLAMA/trec_results_dl20_top100.txt",
        ],
        "md5": "5b2807d1d6f8f7f40cc7b297f9fbdcb6",
        "size (bytes)": 1781403,
        "downloaded": False,
    },
    "SPLADE_P_P_ENSEMBLE_DISTIL/retrieve_results_covid_top100.jsonl": {
        "description": "Top-100 SPLADE++ CoCondenser-EnsembleDistil Results for the TREC COVID task. (Cached JSONL format)",
        "urls": [
            "https://github.com/castorini/rank_llm_data/raw/main/retrieve_results/SPLADE_P_P_ENSEMBLE_DISTIL/retrieve_results_covid_top100.jsonl",
        ],
        "md5": "74a41136fab99a77e8919ac29a6cc659",
        "size (bytes)": 7648893,
        "downloaded": False,
    },
    "SPLADE_P_P_ENSEMBLE_DISTIL/retrieve_results_dl19_top100.jsonl": {
        "description": "Top-100 SPLADE++ CoCondenser-EnsembleDistil Results for the TREC DL19 task. (Cached JSONL format)",
        "urls": [
            "https://github.com/castorini/rank_llm_data/raw/main/retrieve_results/SPLADE_P_P_ENSEMBLE_DISTIL/retrieve_results_dl19_top100.jsonl",
        ],
        "md5": "894d37a5701bde517f75e117b2d5cad4",
        "size (bytes)": 1844101,
        "downloaded": False,
    },
    "SPLADE_P_P_ENSEMBLE_DISTIL/retrieve_results_dl20_top100.jsonl": {
        "description": "Top-100 SPLADE++ CoCondenser-EnsembleDistil Results for the TREC DL20 task. (Cached JSONL format)",
        "urls": [
            "https://github.com/castorini/rank_llm_data/raw/main/retrieve_results/SPLADE_P_P_ENSEMBLE_DISTIL/retrieve_results_dl20_top100.jsonl",
        ],
        "md5": "241c63eba8f4ce8b7795fcaaf565bb03",
        "size (bytes)": 2285466,
        "downloaded": False,
    },
    "SPLADE_P_P_ENSEMBLE_DISTIL/retrieve_results_dl21_top100.jsonl": {
        "description": "Top-100 SPLADE++ CoCondenser-EnsembleDistil Results for the TREC DL21 task. (Cached JSONL format)",
        "urls": [
            "https://github.com/castorini/rank_llm_data/raw/main/retrieve_results/SPLADE_P_P_ENSEMBLE_DISTIL/retrieve_results_dl21_top100.jsonl",
        ],
        "md5": "5b90d5460014d94d977921202c42606a",
        "size (bytes)": 2082006,
        "downloaded": False,
    },
    "SPLADE_P_P_ENSEMBLE_DISTIL/retrieve_results_dl22_top100.jsonl": {
        "description": "Top-100 SPLADE++ CoCondenser-EnsembleDistil Results for the TREC DL22 task. (Cached JSONL format)",
        "urls": [
            "https://github.com/castorini/rank_llm_data/raw/main/retrieve_results/SPLADE_P_P_ENSEMBLE_DISTIL/retrieve_results_dl22_top100.jsonl",
        ],
        "md5": "9082402af60b5e9208556853bc7842b0",
        "size (bytes)": 2958535,
        "downloaded": False,
    },
    "SPLADE_P_P_ENSEMBLE_DISTIL/retrieve_results_news_top100.jsonl": {
        "description": "Top-100 SPLADE++ CoCondenser-EnsembleDistil Results for the TREC News task. (Cached JSONL format)",
        "urls": [
            "https://github.com/castorini/rank_llm_data/raw/main/retrieve_results/SPLADE_P_P_ENSEMBLE_DISTIL/retrieve_results_news_top100.jsonl",
        ],
        "md5": "adcff66eb90b5f66e7ae47af7012fdb0",
        "size (bytes)": 26116086,
        "downloaded": False,
    },
    "SPLADE_P_P_ENSEMBLE_DISTIL/trec_results_covid_top100.txt": {
        "description": "Top-100 SPLADE++ CoCondenser-EnsembleDistil Results for the TREC COVID task. (TREC format)",
        "urls": [
            "https://github.com/castorini/rank_llm_data/raw/main/retrieve_results/retrieve_results/SPLADE_P_P_ENSEMBLE_DISTIL/trec_results_covid_top100.txt",
        ],
        "md5": "d4c49f38e7dff7dd64f3cbed7775e2fc",
        "size (bytes)": 168645,
        "downloaded": False,
    },
    "SPLADE_P_P_ENSEMBLE_DISTIL/trec_results_dl19_top100.txt": {
        "description": "Top-100 SPLADE++ CoCondenser-EnsembleDistil Results for the TREC DL19 task. (TREC format)",
        "urls": [
            "https://github.com/castorini/rank_llm_data/raw/main/retrieve_results/retrieve_results/SPLADE_P_P_ENSEMBLE_DISTIL/trec_results_dl19_top100.txt",
        ],
        "md5": "0fa0c937363726dc48ee675ad53c0221",
        "size (bytes)": 147985,
        "downloaded": False,
    },
    "SPLADE_P_P_ENSEMBLE_DISTIL/trec_results_dl20_top100.txt": {
        "description": "Top-100 SPLADE++ CoCondenser-EnsembleDistil Results for the TREC DL20 task. (TREC format)",
        "urls": [
            "https://github.com/castorini/rank_llm_data/raw/main/retrieve_results/retrieve_results/SPLADE_P_P_ENSEMBLE_DISTIL/trec_results_dl20_top100.txt",
        ],
        "md5": "323443d55fc9c50698d88288f53ffe8d",
        "size (bytes)": 186223,
        "downloaded": False,
    },
    "SPLADE_P_P_ENSEMBLE_DISTIL/trec_results_dl21_top100.txt": {
        "description": "Top-100 SPLADE++ CoCondenser-EnsembleDistil Results for the TREC DL21 task. (TREC format)",
        "urls": [
            "https://github.com/castorini/rank_llm_data/raw/main/retrieve_results/retrieve_results/SPLADE_P_P_ENSEMBLE_DISTIL/trec_results_dl21_top100.txt",
        ],
        "md5": "dcb67a2c9cca231c28711abecb258c8a",
        "size (bytes)": 312606,
        "downloaded": False,
    },
    "SPLADE_P_P_ENSEMBLE_DISTIL/trec_results_dl22_top100.txt": {
        "description": "Top-100 SPLADE++ CoCondenser-EnsembleDistil Results for the TREC DL22 task. (TREC format)",
        "urls": [
            "https://github.com/castorini/rank_llm_data/raw/main/retrieve_results/retrieve_results/SPLADE_P_P_ENSEMBLE_DISTIL/trec_results_dl22_top100.txt",
        ],
        "md5": "8daf6f01dd39d73b9e33bf218c3b6c38",
        "size (bytes)": 454452,
        "downloaded": False,
    },
    "SPLADE_P_P_ENSEMBLE_DISTIL/trec_results_news_top100.txt": {
        "description": "Top-100 SPLADE++ CoCondenser-EnsembleDistil Results for the TREC News task. (TREC format)",
        "urls": [
            "https://github.com/castorini/rank_llm_data/raw/main/retrieve_results/retrieve_results/SPLADE_P_P_ENSEMBLE_DISTIL/trec_results_news_top100.txt",
        ],
        "md5": "d67ff0cae7048e9cd9183c788d804296",
        "size (bytes)": 343746,
        "downloaded": False,
    },
}



import time
from enum import Enum
from typing import Any, Dict, List, Optional, Tuple, Union

import openai
import tiktoken
from tqdm import tqdm

from rank_llm.data import Request, Result
from rank_llm.rerank import PromptMode

from .listwise_rankllm import ListwiseRankLLM


class SafeOpenai(ListwiseRankLLM):
    def __init__(
        self,
        model: str,
        context_size: int,
        prompt_mode: PromptMode = PromptMode.RANK_GPT,
        num_few_shot_examples: int = 0,
        window_size: int = 20,
        keys=None,
        key_start_id=None,
        proxy=None,
        api_type: str = None,
        api_base: str = None,
        api_version: str = None,
    ) -> None:
        """
        Creates instance of the SafeOpenai class, a specialized version of RankLLM designed for safely handling OpenAI API calls with
        support for key cycling, proxy configuration, and Azure AI conditional integration.

        Parameters:
        - model (str): The model identifier for the LLM (model identifier information can be found via OpenAI's model lists).
        - context_size (int): The maximum number of tokens that the model can handle in a single request.
        - prompt_mode (PromptMode, optional): Specifies the mode of prompt generation, with the default set to RANK_GPT,
         indicating that this class is designed primarily for listwise ranking tasks following the RANK_GPT methodology.
        - num_few_shot_examples (int, optional): Number of few-shot learning examples to include in the prompt, allowing for
        the integration of example-based learning to improve model performance. Defaults to 0, indicating no few-shot examples
        by default.
        - window_size (int, optional): The window size for handling text inputs. Defaults to 20.
        - keys (Union[List[str], str], optional): A list of OpenAI API keys or a single OpenAI API key.
        - key_start_id (int, optional): The starting index for the OpenAI API key cycle.
        - proxy (str, optional): The proxy configuration for OpenAI API calls.
        - api_type (str, optional): The type of API service, if using Azure AI as the backend.
        - api_base (str, optional): The base URL for the API, applicable when using Azure AI.
        - api_version (str, optional): The API version, necessary for Azure AI integration.

        Raises:
        - ValueError: If an unsupported prompt mode is provided or if no OpenAI API keys / invalid OpenAI API keys are supplied.

        Note:
        - This class supports cycling between multiple OpenAI API keys to distribute quota usage or handle rate limiting.
        - Azure AI integration is depends on the presence of `api_type`, `api_base`, and `api_version`.
        """
        super().__init__(
            model, context_size, prompt_mode, num_few_shot_examples, window_size
        )
        if isinstance(keys, str):
            keys = [keys]
        if not keys:
            raise ValueError("Please provide OpenAI Keys.")
        if prompt_mode not in [
            PromptMode.RANK_GPT,
            PromptMode.RANK_GPT_APEER,
            PromptMode.LRL,
        ]:
            raise ValueError(
                f"unsupported prompt mode for GPT models: {prompt_mode}, expected {PromptMode.RANK_GPT}, {PromptMode.RANK_GPT_APEER} or {PromptMode.LRL}."
            )

        self._output_token_estimate = None
        self._keys = keys
        self._cur_key_id = key_start_id or 0
        self._cur_key_id = self._cur_key_id % len(self._keys)
        openai.proxy = proxy
        openai.api_key = self._keys[self._cur_key_id]
        self.use_azure_ai = False

        if all([api_type, api_base, api_version]):
            # See https://learn.microsoft.com/en-US/azure/ai-services/openai/reference for list of supported versions
            openai.api_version = api_version
            openai.api_type = api_type
            openai.api_base = api_base
            self.use_azure_ai = True

    class CompletionMode(Enum):
        UNSPECIFIED = 0
        CHAT = 1
        TEXT = 2

    def rerank_batch(
        self,
        requests: List[Request],
        rank_start: int = 0,
        rank_end: int = 100,
        shuffle_candidates: bool = False,
        logging: bool = False,
        **kwargs: Any,
    ) -> List[Result]:
        top_k_retrieve: int = kwargs.get("top_k_retrieve", rank_end)
        rank_end = min(top_k_retrieve, rank_end)
        window_size: int = kwargs.get("window_size", 20)
        window_size = min(window_size, top_k_retrieve)
        step: int = kwargs.get("step", 10)
        populate_invocations_history: bool = kwargs.get(
            "populate_invocations_history", False
        )
        results = []
        for request in tqdm(requests):
            result = self.sliding_windows(
                request,
                rank_start=max(rank_start, 0),
                rank_end=min(rank_end, len(request.candidates)),
                window_size=window_size,
                step=step,
                shuffle_candidates=shuffle_candidates,
                logging=logging,
                populate_invocations_history=populate_invocations_history,
            )
            results.append(result)
        return results

    def _call_completion(
        self,
        *args,
        completion_mode: CompletionMode,
        return_text=False,
        reduce_length=False,
        **kwargs,
    ) -> Union[str, Dict[str, Any]]:
        while True:
            try:
                if completion_mode == self.CompletionMode.CHAT:
                    completion = openai.chat.completions.create(
                        *args, **kwargs, timeout=30
                    )
                elif completion_mode == self.CompletionMode.TEXT:
                    completion = openai.Completion.create(*args, **kwargs)
                else:
                    raise ValueError(
                        "Unsupported completion mode: %V" % completion_mode
                    )
                break
            except Exception as e:
                print("Error in completion call")
                print(str(e))
                if "This model's maximum context length is" in str(e):
                    print("reduce_length")
                    return "ERROR::reduce_length"
                if "The response was filtered" in str(e):
                    print("The response was filtered")
                    return "ERROR::The response was filtered"
                self._cur_key_id = (self._cur_key_id + 1) % len(self._keys)
                openai.api_key = self._keys[self._cur_key_id]
                time.sleep(0.1)
        if return_text:
            completion = (
                completion.choices[0].message.content
                if completion_mode == self.CompletionMode.CHAT
                else completion.choices[0].text
            )
        return completion

    def run_llm(
        self,
        prompt: Union[str, List[Dict[str, str]]],
        current_window_size: Optional[int] = None,
    ) -> Tuple[str, int]:
        model_key = "model"
        response = self._call_completion(
            messages=prompt,
            temperature=0,
            completion_mode=SafeOpenai.CompletionMode.CHAT,
            return_text=True,
            **{model_key: self._model},
        )
        try:
            encoding = tiktoken.get_encoding(self._model)
        except:
            encoding = tiktoken.get_encoding("cl100k_base")
        return response, len(encoding.encode(response))

    def _get_prefix_for_rank_gpt_prompt(
        self, query: str, num: int
    ) -> List[Dict[str, str]]:
        return [
            {
                "role": "system",
                "content": "You are RankGPT, an intelligent assistant that can rank passages based on their relevancy to the query.",
            },
            {
                "role": "user",
                "content": f"I will provide you with {num} passages, each indicated by number identifier []. \nRank the passages based on their relevance to query: {query}.",
            },
            {"role": "assistant", "content": "Okay, please provide the passages."},
        ]

    def _get_suffix_for_rank_gpt_prompt(self, query: str, num: int) -> str:
        return f"Search Query: {query}. \nRank the {num} passages above based on their relevance to the search query. The passages should be listed in descending order using identifiers. The most relevant passages should be listed first. The output format should be [] > [], e.g., [1] > [2]. Only response the ranking results, do not say any word or explain."

    def _get_prefix_for_rank_gpt_apeer_prompt(
        self, query: str, num: int
    ) -> List[Dict[str, str]]:
        # APEER
        return [
            {
                "role": "system",
                "content": "As RankGPT, your task is to evaluate and rank unique passages based on their relevance and accuracy to a given query. Prioritize passages that directly address the query and provide detailed, correct answers. Ignore factors such as length, complexity, or writing style unless they seriously hinder readability.",
            },
            {
                "role": "user",
                "content": f"In response to the query: [querystart] {query} [queryend], rank the passages. Ignore aspects like length, complexity, or writing style, and concentrate on passages that provide a comprehensive understanding of the query. Take into account any inaccuracies or vagueness in the passages when determining their relevance.",
            },
        ]

    def _get_suffix_for_rank_gpt_apeer_prompt(self, query: str, num: int) -> str:
        return f"Given the query: [querystart] {query} [queryend], produce a succinct and clear ranking of all passages, from most to least relevant, using their identifiers. The format should be [rankstart] [most relevant passage ID] > [next most relevant passage ID] > ... > [least relevant passage ID] [rankend]. Refrain from including any additional commentary or explanations in your ranking."

    def num_output_tokens(self, current_window_size: Optional[int] = None) -> int:
        if current_window_size is None:
            current_window_size = self._window_size
        if self._output_token_estimate and self._window_size == current_window_size:
            return self._output_token_estimate
        else:
            try:
                encoder = tiktoken.get_encoding(self._model)
            except:
                encoder = tiktoken.get_encoding("cl100k_base")

            _output_token_estimate = (
                len(
                    encoder.encode(
                        " > ".join([f"[{i+1}]" for i in range(current_window_size)])
                    )
                )
                - 1
            )
            if (
                self._output_token_estimate is None
                and self._window_size == current_window_size
            ):
                self._output_token_estimate = _output_token_estimate
            return _output_token_estimate

    def create_prompt_batched(self):
        pass

    def run_llm_batched(self):
        pass

    def create_prompt(
        self, result: Result, rank_start: int, rank_end: int
    ) -> Tuple[List[Dict[str, str]], int]:
        if self._prompt_mode in [PromptMode.RANK_GPT, PromptMode.RANK_GPT_APEER]:
            return self.create_rank_gpt_prompt(result, rank_start, rank_end)
        else:
            return self.create_LRL_prompt(result, rank_start, rank_end)

    def create_rank_gpt_prompt(
        self, result: Result, rank_start: int, rank_end: int
    ) -> Tuple[List[Dict[str, str]], int]:
        query = result.query.text
        num = len(result.candidates[rank_start:rank_end])

        max_length = 300 * (self._window_size / (rank_end - rank_start))
        while True:
            messages = (
                self._get_prefix_for_rank_gpt_apeer_prompt(query, num)
                if self._prompt_mode == PromptMode.RANK_GPT_APEER
                else self._get_prefix_for_rank_gpt_prompt(query, num)
            )
            rank = 0
            for cand in result.candidates[rank_start:rank_end]:
                rank += 1
                content = self.convert_doc_to_prompt_content(cand.doc, max_length)
                if self._prompt_mode == PromptMode.RANK_GPT_APEER:
                    messages[-1][
                        "content"
                    ] += f"\n[{rank}] {self._replace_number(content)}"
                else:
                    messages.append(
                        {
                            "role": "user",
                            "content": f"[{rank}] {self._replace_number(content)}",
                        }
                    )
                    messages.append(
                        {"role": "assistant", "content": f"Received passage [{rank}]."}
                    )
            if self._prompt_mode == PromptMode.RANK_GPT_APEER:
                messages[-1][
                    "content"
                ] += f"\n{self._get_suffix_for_rank_gpt_apeer_prompt(query, num)}"
            else:
                messages.append(
                    {
                        "role": "user",
                        "content": self._get_suffix_for_rank_gpt_prompt(query, num),
                    }
                )
            num_tokens = self.get_num_tokens(messages)
            if num_tokens <= self.max_tokens() - self.num_output_tokens():
                break
            else:
                max_length -= max(
                    1,
                    (num_tokens - self.max_tokens() + self.num_output_tokens())
                    // ((rank_end - rank_start) * 4),
                )
        return messages, self.get_num_tokens(messages)

    def create_LRL_prompt(
        self, result: Result, rank_start: int, rank_end: int
    ) -> Tuple[List[Dict[str, str]], int]:
        query = result.query.text
        num = len(result.candidates[rank_start:rank_end])
        max_length = 300 * (20 / (rank_end - rank_start))
        psg_ids = []
        while True:
            message = "Sort the list PASSAGES by how good each text answers the QUESTION (in descending order of relevancy).\n"
            rank = 0
            for cand in result.candidates[rank_start:rank_end]:
                rank += 1
                psg_id = f"PASSAGE{rank}"
                content = self.convert_doc_to_prompt_content(cand.doc, max_length)
                message += f'{psg_id} = "{self._replace_number(content)}"\n'
                psg_ids.append(psg_id)
            message += f'QUESTION = "{query}"\n'
            message += "PASSAGES = [" + ", ".join(psg_ids) + "]\n"
            message += "Sort the PASSAGES by their relevance to the Query. The answer should be a sorted list of PASSAGE ids (e.g., [PASSAGE2, ..., PASSAGE1]). Do not include any additional words or explanations.\n"
            message += "SORTED_PASSAGES = "
            messages = [{"role": "user", "content": message}]
            num_tokens = self.get_num_tokens(messages)
            if num_tokens <= self.max_tokens() - self.num_output_tokens():
                break
            else:
                max_length -= max(
                    1,
                    (num_tokens - self.max_tokens() + self.num_output_tokens())
                    // ((rank_end - rank_start) * 4),
                )
        return messages, self.get_num_tokens(messages)

    def get_num_tokens(self, prompt: Union[str, List[Dict[str, str]]]) -> int:
        """Returns the number of tokens used by a list of messages in prompt."""
        if self._model in ["gpt-3.5-turbo-0301", "gpt-3.5-turbo"]:
            tokens_per_message = (
                4  # every message follows <|start|>{role/name}\n{content}<|end|>\n
            )
            tokens_per_name = -1  # if there's a name, the role is omitted
        elif self._model in ["gpt-4-0314", "gpt-4"]:
            tokens_per_message = 3
            tokens_per_name = 1
        else:
            tokens_per_message, tokens_per_name = 0, 0

        try:
            encoding = tiktoken.get_encoding(self._model)
        except:
            encoding = tiktoken.get_encoding("cl100k_base")

        num_tokens = 0
        if isinstance(prompt, list):
            for message in prompt:
                num_tokens += tokens_per_message
                for key, value in message.items():
                    num_tokens += len(encoding.encode(value))
                    if key == "name":
                        num_tokens += tokens_per_name
        else:
            num_tokens += len(encoding.encode(prompt))
        num_tokens += 3  # every reply is primed with <|start|>assistant<|message|>
        return num_tokens

    def cost_per_1k_token(self, input_token: bool) -> float:
        # Brought in from https://openai.com/pricing on 2023-07-30
        cost_dict = {
            ("gpt-3.5", 4096): 0.0015 if input_token else 0.002,
            ("gpt-3.5", 16384): 0.003 if input_token else 0.004,
            ("gpt-4", 8192): 0.03 if input_token else 0.06,
            ("gpt-4", 32768): 0.06 if input_token else 0.12,
        }
        model_key = "gpt-3.5" if "gpt-3" in self._model else "gpt-4"
        return cost_dict[(model_key, self._context_size)]

    def get_name(self) -> str:
        return self._model
